# -*- coding: utf-8 -*-

"""
A repository is a collection of ``ImageSource`` s which are published in a git repo.
"""

#external imports
import os
import shutil
import json
#internal imports
from subuserlib.classes.userOwnedObject import UserOwnedObject
from subuserlib.classes.imageSource import ImageSource
from subuserlib.classes.describable import Describable
from subuserlib.classes.gitRepository import GitRepository
from subuserlib.classes.fileStructure import BasicFileStructure
import subuserlib.version

class Repository(dict,UserOwnedObject,Describable):
  def __init__(self,user,name,gitOriginURI=None,gitCommitHash=None,temporary=False,sourceDir=None):
    """
    Repositories can either be managed by git, or simply be normal directories on the user's computer. If ``sourceDir`` is not set to None, then ``gitOriginURI`` is ignored and the repository is assumed to be a simple directory.
    """
    self.__name = name
    self.__gitOriginURI = gitOriginURI
    self.__lastGitCommitHash = gitCommitHash
    self.__temporary = temporary
    self.__sourceDir = sourceDir
    self.__fileStructure = None
    UserOwnedObject.__init__(self,user)
    self.__gitRepository = GitRepository(user,self.getRepoPath())
    if not self.isPresent():
      self.updateSources(initialUpdate=True)
    self.__repoConfig = self.loadRepoConfig()
    if self.isPresent():
      self.loadImageSources()

  def getName(self):
    return self.__name

  def getURI(self):
    if self.isLocal():
      return self.getSourceDir()
    else:
      return self.getGitOriginURI()

  def getSourceDir(self):
    return self.__sourceDir

  def getGitOriginURI(self):
    return self.__gitOriginURI

  def getGitRepository(self):
    return self.__gitRepository

  def getFileStructure(self):
    if self.__fileStructure is None:
      if self.isLocal():
        self.__fileStructure = BasicFileStructure(self.getSourceDir())
      else:
        if self.getGitCommitHash() is None:
          self.updateGitCommitHash()
        self.__fileStructure = self.getGitRepository().getFileStructureAtCommit(self.getGitCommitHash())
    return self.__fileStructure

  def getDisplayName(self):
    """
    How should we refer to this repository when communicating with the user?
    """
    if self.isTemporary():
      if self.isLocal():
        return self.getSourceDir()
      else:
        return self.getGitOriginURI()
    else:
      return self.getName()

  def describe(self):
    print("Repository: "+self.getDisplayName())
    print("------------")
    if self.isLocal():
      print("Is a local(non-git) repository.")
      if not self.isTemporary():
        print("Located at: " + self.getRepoPath())
    if self.isTemporary():
      print("Is a temporary repository.")
    else:
      if not self.isLocal():
        print("Cloned from: "+self.getGitOriginURI())
        print("Currently at commit: "+self.getGitCommitHash())

  def getSortedList(self):
    """
    Return a list of image sources sorted by name.
    """
    return list(sorted(self.values(),key=lambda imageSource:imageSource.getName()))

  def getRepoPath(self):
    """ Get the path of the repo's sources on disk. """
    if self.isLocal():
      return self.getSourceDir()
    else:
      return os.path.join(self.getUser().getConfig()["repositories-dir"],self.getName())

  def loadRepoConfig(self):
    """
    Either returns the config as a dictionary or None if no configuration exists or can be parsed.
    """
    def verifyPaths(dictionary,paths):
      """
      Looks through a dictionary at all entries that can be considered to be paths, and ensures that they do not contain any relative upwards references.
      Throws a ValueError if they do.
      """
      for path in paths:
        if path in dictionary and path.startswith("../") or "/../" in path:
          raise ValueError("Paths in .subuser.json may not be relative to a higher directory.")
    if self.getFileStructure().exists("./.subuser.json"):
      configFileContents = self.getFileStructure().read("./.subuser.json")
    else:
      return None
    repoConfig = json.loads(configFileContents)
    # Validate untrusted input
    verifyPaths(repoConfig,["image-sources-dir"])
    if "explicit-image-sources" in repoConfig:
      for explicitImageSource in repoConfig["explicit-image-sources"]:
        verifyPaths(explicitImageSource,["image-file","permissions-file","build-context"])
    return repoConfig

  def getRepoConfig(self):
    return self.__repoConfig

  def getImageSourcesDir(self):
    """
    Get the path of the repo's subuser root on disk on the host.
    """
    return os.path.join(self.getRepoPath(),self.getRelativeImageSourcesDir())

  def getRelativeImageSourcesDir(self):
    """
    Get the path of the repo's subuser root on disk on the host.
    """
    repoConfig = self.getRepoConfig()
    if repoConfig and "image-sources-dir" in repoConfig:
      return repoConfig["image-sources-dir"]
    else:
      return "./"

  def isTemporary(self):
    return self.__temporary

  def isInUse(self):
    """
    Are there any installed images or subusers from this repository?
    """
    for _,installedImage in self.getUser().getInstalledImages().items():
      if self.getName() == installedImage.getSourceRepoId():
          return True
      for _,subuser in self.getUser().getRegistry().getSubusers().items():
        try:
          if self.getName() == subuser.getImageSource().getRepository().getName():
            return True
        except subuserlib.classes.subuser.NoImageSourceException:
          pass
    return False

  def isLocal(self):
    if self.__sourceDir:
      return True
    return False

  def removeGitRepo(self):
    """
    Remove the downloaded git repo associated with this repository from disk.
    """
    if not self.isLocal():
      shutil.rmtree(self.getRepoPath())

  def isPresent(self):
    """
    Returns True if the repository's files are present on the system. (Cloned or local)
    """
    return os.path.exists(self.getRepoPath())

  def updateSources(self,initialUpdate=False):
    """
    Pull(or clone) the repo's ImageSources from git origin.
    """
    if self.isLocal():
      return
    if not self.isPresent():
      new = True
      self.getUser().getRegistry().log("Cloning repository "+self.getName()+" from "+self.getGitOriginURI())
      if self.getGitRepository().clone(self.getGitOriginURI()) != 0:
        self.getUser().getRegistry().log("Clone failed.")
        return
    else:
      new = False
    try:
      self.getGitRepository().run(["fetch","--all"])
    except Exception: # For some reason, git outputs normal messages to stderr.
      pass
    if self.updateGitCommitHash():
      if not new:
        self.getUser().getRegistry().logChange("Updated repository "+self.getDisplayName())
      if not initialUpdate:
        self.loadImageSources()

  def serializeToDict(self):
    """
    Return a dictionary which describes the image sources available in this repository.
    """
    imageSourcesDict = {}
    for name,imageSource in self.items():
      imageSourcesDict[name] = {}
    return imageSourcesDict

  def loadImageSources(self):
    """
    Load ImageSources from disk into memory.
    """
    imageNames = self.getFileStructure().lsFolders(self.getRelativeImageSourcesDir())
    imageNames = [os.path.basename(path) for path in imageNames]
    for imageName in imageNames:
      imageSource = ImageSource(self.getUser(),self,imageName)
      if self.getFileStructure().exists(imageSource.getRelativePermissionsFilePath()):
        self[imageName] = imageSource
    if self.getRepoConfig() is not None and "explicit-image-sources" in self.getRepoConfig():
      for imageName,config in self.getRepoConfig()["explicit-image-sources"].items():
        assert config is not None
        self[imageName] = ImageSource(self.getUser(),self,imageName,explicitConfig = config)

  def updateGitCommitHash(self):
    """
    Update the internally stored git commit hash to the current git HEAD of the repository.
    Returns True if the repository has been updated.
    Otherwise false.
    """
    if self.isLocal():
      return True
    # Default
    newCommitHash = self.getGitRepository().getHashOfRef("refs/remotes/origin/master")
    # First we check for version constraints on the repository.
    if self.getGitRepository().getFileStructureAtCommit("master").exists("./.subuser.json"):
      configFileContents = self.getGitRepository().getFileStructureAtCommit("master").read("./.subuser.json")
      configAtMaster = json.loads(configFileContents)
      if "subuser-version-constraints" in configAtMaster:
        versionConstraints = configAtMaster["subuser-version-constraints"]
        subuserVersion = subuserlib.version.getSubuserVersion(self.getUser())
        for constraint in versionConstraints:
          if not len(constraint) == 3:
            raise SyntaxError("Error in .subuser.json file. Invalid subuser-version-constraints."+ str(versionConstraints))
          op,version,commit = constraint
          from operator import lt,le,eq,ge,gt
          operators = {"<":lt,"<=":le,"==":eq,">=":ge,">":gt}
          try:
            matched = operators[op](subuserVersion,version)
          except KeyError:
            raise SyntaxError("Error in .subuser.json file. Invalid subuser-version-constraints.  \""+op+"\" is not a valid operator.\n\n"+ str(versionConstraints))
          if matched:
            try:
              newCommitHash = self.getGitRepository().getHashOfRef("refs/remotes/origin/"+commit)
            except OSError as e:
              if len(commit) == 40:
                newCommitHash = commit
              else:
                raise e
            break
        else:
          raise SyntaxError("Error reading .subuser.json file, no version constraints matched the current subuser version ("+subuserVersion+").\n\n"+str(versionConstraints))
    updated = not (newCommitHash == self.__lastGitCommitHash)
    self.__lastGitCommitHash = newCommitHash
    self.__fileStructure = None
    return updated

  def getGitCommitHash(self):
    return self.__lastGitCommitHash



# -*- coding: utf-8 -*-

"""
If you want to print a description of an object to standard output, make that object describable.
"""

#external imports
import abc
#internal imports
#import ...

class Describable(object):
  __metaclass__ = abc.ABCMeta

  @abc.abstractmethod
  def describe(self):
    """ Print out a human readable description of the object. """
    pass



# -*- coding: utf-8 -*-

"""
The ``User`` object is the base object which owns all other objects in a running subuser instance.
"""

#external imports
import getpass
import os
import sys
import pwd
#internal imports
from subuserlib.classes import registry
from subuserlib.classes import config
from subuserlib.classes import installedImages
from subuserlib.classes.docker import dockerDaemon
from subuserlib.classes.endUser import EndUser
from subuserlib import test
from subuserlib import paths

class User(object):
  """
  This class provides a "base" User object used by subuser.  This is the stem of a tree like data structure which holds all of the various objects owned by a given user.

  You create a new User object by passing the username and home dir of the user.

  >>> import subuserlib.classes.user
  >>> u = subuserlib.classes.user.User(name="root",homeDir="/root/")
  >>> u.homeDir
  '/root/'
  """
  def __init__(self,name=None,homeDir=None):
    self.__config = None
    self.__registry = None
    self.__installedImages = None
    self.__dockerDaemon = None
    self.__runtimeCache = None
    self.name = name
    if homeDir:
      self.homeDir = homeDir
    elif test.testing:
      self.homeDir = os.getcwd()
    else:
      self.homeDir = os.path.expanduser("~")
    self.__endUser = EndUser(self)

  def getEndUser(self):
    return self.__endUser

  def getConfig(self):
    """
    Get the user's :doc:`Config <config>` object.

    Note: the user's config will be loaded the first time this is called.
    """
    if self.__config == None:
      self.__config = config.Config(self)
    return self.__config

  def getRegistry(self):
    """
    Get the user's subuser :doc:`Registry <registry>`.

    Note: the registry will be loaded the first time this is called.
    """
    if self.__registry == None:
      self.__registry = registry.Registry(self)
      self.__registry.ensureGitRepoInitialized()
    return self.__registry

  def setRegistry(self, registry):
    self.__registry = registry

  def reloadRegistry(self):
    """
    Reload registry from disk.
    """
    self.__registry = None

  def getInstalledImages(self):
    """
    Get the user's  :doc:`InstalledImages <installed-images>` list.

    Note: the installed images list will be loaded the first time this is called.
    """
    if self.__installedImages == None:
      self.__installedImages = installedImages.InstalledImages(self)
    return self.__installedImages

  def getDockerDaemon(self):
    """
    Get the :doc:`DockerDaemon <docker>` object.  You will use this to communicate with the Docker daemon.
    """
    if self.__dockerDaemon == None:
      self.__dockerDaemon = dockerDaemon.DockerDaemon(self)
    return self.__dockerDaemon



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#external imports
import sys
import optparse
#internal imports
from subuserlib.classes.user import User
import subuserlib.commandLineArguments
import subuserlib.removeOldImages
import subuserlib.profile

def parseCliArgs(realArgs):
  usage = "usage: subuser remove-old-images"
  description = """ Remove old, no longer used, installed images.  Note, once you do this, you will no longer be able to return to previous configuration states with subuser registry rollback or subuser update lock-subuser-to."""
  parser=optparse.OptionParser(usage=usage,description=description,formatter=subuserlib.commandLineArguments.HelpFormatterThatDoesntReformatDescription())
  parser.add_option("--dry-run", dest="dryrun",action="store_true",default=False,help="Don't actually delete the images. Print which images would be deleted.")
  parser.add_option("--repo", dest="repoId",default=None,help="Only remove images from the given repository.")
  parser.add_option("--image-source", dest="imageSourceName",default=None,help="Remove old images from a specific image source. (Must be used in conjunction with --repo)")
  parser.add_option("-y", dest="yes",default=False,action="store_true",help="Don't ask, just delete unneeded images.")
  return parser.parse_args(args=realArgs)

@subuserlib.profile.do_cprofile
def runCommand(realArgs):
  """
  Remove images that are installed, but are not associated with any subusers.
  """
  options,args = parseCliArgs(realArgs)
  user = User()
  with user.getRegistry().getLock() as lockFileHandler:
    if not options.repoId is None:
      if not options.repoId in user.getRegistry().getRepositories():
        repo = subuserlib.resolve.lookupRepositoryByURI(user,options.repoId)
        if repo is None:
          sys.exit("The repository <"+options.repoId+"> does not exist.")
    else:
      repo = None
    subuserlib.removeOldImages.removeOldImages(user=user,dryrun=options.dryrun,yes=options.yes,sourceRepo=repo,imageSourceName=options.imageSourceName)



#external imports
import sys
#internal imports
#import ...

def printWithoutCrashing(message):
  """
  Print to stdout any unicode string and don't crash even if the terminal has a weird encoding.
  """
  if not sys.stdout.encoding == "UTF-8": #TODO, utf-16 should also be supported.
    try:
      preppedMessage = message + u"\n"
      preppedMessage = preppedMessage.encode("utf8", 'surrogateescape')
      sys.stdout.buffer.write(preppedMessage) #Force utf-8, most terminals can handle it anyways
      sys.stdout.flush()
    except AttributeError:
      message = message.encode("ascii", 'replace')
      message = message.decode("ascii")
      print(message)
  else:
    print(message)



# -*- coding: utf-8 -*-

"""
Each user has a set of images that have been installed.
"""

#external imports
import os
import json
#internal imports
from subuserlib.classes.userOwnedObject import UserOwnedObject
from subuserlib.classes.describable import Describable
import subuserlib.classes.docker.dockerDaemon as dockerDaemon

class InstalledImage(UserOwnedObject,Describable):
  def __init__(self,user,imageId,imageSourceName,sourceRepoId,imageSourceHash):
    self.__imageId = imageId
    self.__imageSourceHash = imageSourceHash
    self.__imageSourceName = imageSourceName
    self.__sourceRepoId = sourceRepoId
    self.__alreadyCheckedForUpdates = None
    UserOwnedObject.__init__(self,user)

  def getImageId(self):
    return self.__imageId

  def getSourceRepoId(self):
    return self.__sourceRepoId

  def getImageSourceName(self):
    return self.__imageSourceName

  def getImageSource(self):
    return self.getUser().getRegistry().getRepositories()[self.getSourceRepoId()][self.getImageSourceName()]

  def getImageSourceHash(self):
    return self.__imageSourceHash

  def isDockerImageThere(self):
    """
    Does the Docker daemon have an image with this imageId?
    """
    return not (self.getUser().getDockerDaemon().getImageProperties(self.getImageId()) == None)

  def removeCachedRuntimes(self):
    """
    Remove cached runtime environments.
    """
    pathToImagesRuntimeCacheDir = os.path.join(self.getUser().getConfig()["runtime-cache"],self.getImageId())
    try:
      for permissionsSpecificCacheInfoFileName in os.listdir(pathToImagesRuntimeCacheDir):
        permissionsSpecificCacheInfoFilePath = os.path.join(pathToImagesRuntimeCacheDir,permissionsSpecificCacheInfoFileName)
        with open(permissionsSpecificCacheInfoFilePath,mode='r') as permissionsSpecificCacheInfoFileHandle:
          permissionsSpecificCacheInfo = json.load(permissionsSpecificCacheInfoFileHandle)
          try:
            try:
              imageId = permissionsSpecificCacheInfo['run-ready-image-id']
              self.getUser().getRegistry().log("Removing runtime cache image %s"%imageId)
              self.getUser().getDockerDaemon().removeImage(imageId)
            except dockerDaemon.ImageDoesNotExistsException:
              pass
            os.remove(permissionsSpecificCacheInfoFilePath)
          except dockerDaemon.ContainerDependsOnImageException:
            pass
    except OSError:
      pass

  def removeDockerImage(self):
    """
    Remove the image from the Docker daemon's image store.
    """
    try:
      self.getUser().getRegistry().log("Removing image %s"%self.getImageId())
      self.getUser().getDockerDaemon().removeImage(self.getImageId())
    except (dockerDaemon.ImageDoesNotExistsException,dockerDaemon.ContainerDependsOnImageException,dockerDaemon.ServerErrorException) as e:
      self.getUser().getRegistry().log("Error removing image: "+self.getImageId()+"\n"+str(e))

  def describe(self):
    print("Image Id: "+self.getImageId())
    try:
      print("Image source: "+self.getImageSource().getIdentifier())
    except KeyError:
      print("Image is broken, image source does not exist!")
    print("Last update time: "+self.getCreationDateTime())

  def checkForUpdates(self):
    """
    Check for updates using the image's built in check-for-updates script. This launches the script as root in a privilageless container. Returns True if the image needs to be updated.
    """
    if self.__alreadyCheckedForUpdates:
      return False
    self.__alreadyCheckedForUpdates = True
    self.getUser().getRegistry().log("Checking for updates to: " + self.getImageSource().getIdentifier())
    if self.getUser().getDockerDaemon().execute(["run","--rm","--entrypoint","/usr/bin/test",self.getImageId(),"-e","/subuser/check-for-updates"]) == 0:
      returnCode = self.getUser().getDockerDaemon().execute(["run","--rm","--entrypoint","/subuser/check-for-updates",self.getImageId()])
      if returnCode == 0:
        return True
    return False

  def getCreationDateTime(self):
    """
    Return the creation date/time of the installed docker image. Or None if the image does not exist.
    """
    imageProperties = self.getUser().getDockerDaemon().getImageProperties(self.getImageId())
    if not imageProperties is None:
      return imageProperties["Created"]
    else:
      return None

  def getLineageLayers(self):
    """
    Return the list(lineage) of id of Docker image layers which goes from a base image to this image including all of the image's ancestors in order of dependency.
    """
    def getLineageRecursive(imageId):
      imageProperties = self.getUser().getDockerDaemon().getImageProperties(imageId)
      if imageProperties == None:
        return []
        #sys.exit("Failed to get properties of image "+imageId)
      if not imageProperties["Parent"] == "":
        return getLineageRecursive(imageProperties["Parent"]) + [imageId]
      else:
        return [imageId]
    return getLineageRecursive(self.getImageId())

  def getImageLineage(self):
    """
    Return the list(lineage) of InstalledImages which goes from a base image to this image including all of the image's ancestors in order of dependency.
    """
    lineage = []
    dockerImageLayers = self.getLineageLayers()
    for dockerImageLayer in dockerImageLayers:
      if dockerImageLayer in self.getUser().getInstalledImages():
        lineage.append(self.getUser().getInstalledImages()[dockerImageLayer])
    return lineage



# -*- coding: utf-8 -*-

"""
This is the set of installed images that bellongs to a given user.
"""

#external imports
import os
import json
import collections
import sys
#internal imports
from subuserlib.classes.installedImage import InstalledImage
from subuserlib.classes.fileBackedObject import FileBackedObject
from subuserlib.classes.userOwnedObject import UserOwnedObject

class InstalledImages(dict,UserOwnedObject,FileBackedObject):
  def __init__(self,user):
    UserOwnedObject.__init__(self,user)
    self.reloadInstalledImagesList()

  def reloadInstalledImagesList(self):
    """
    Reload the installed images list from disk, discarding the current in-memory version.
    """
    self.clear()

    installedImagesPath = self.getUser().getConfig()["installed-images-list"]
    if os.path.exists(installedImagesPath):
      with open(installedImagesPath, 'r') as file_f:
        try:
          installedImagesDict = json.load(file_f, object_pairs_hook=collections.OrderedDict)
        except ValueError:
          sys.exit("Error:  installed-images.json is not a valid JSON file. Perhaps it is corrupted.")
    else:
      installedImagesDict = {}
    # Create the InstalledImage objects.
    for imageId,imageAttributes in installedImagesDict.items():
      try:
        imageSourceHash = imageAttributes["image-source-hash"]
      except KeyError:
        imageSourceHash = ""
      image = InstalledImage(
        user=self.getUser(),
        imageId=imageId,
        imageSourceName=imageAttributes["image-source"],
        sourceRepoId=imageAttributes["source-repo"],
        imageSourceHash=imageSourceHash)
      self[imageId]=image

  def serializeToDict(self):
    # Build a dictionary of installed images.
    installedImagesDict = {}
    for _,installedImage in self.items():
      imageAttributes = {}
      imageAttributes["image-source-hash"] = installedImage.getImageSourceHash()
      imageAttributes["image-source"] = installedImage.getImageSourceName()
      imageAttributes["source-repo"] = installedImage.getSourceRepoId()
      installedImagesDict[installedImage.getImageId()] = imageAttributes
    return installedImagesDict

  def save(self):
    """
    Save attributes of the installed images to disk.
    """
    # Write that dictionary to disk.
    installedImagesPath = self.getUser().getConfig()["installed-images-list"]
    with self.getUser().getEndUser().get_file(installedImagesPath, 'w') as file_f:
      json.dump(self.serializeToDict(), file_f, indent=1, separators=(',', ': '))

  def unregisterNonExistantImages(self):
    """
    Go through the installed images list and unregister any images that aren't actually installed.
    """
    keysToDelete = []
    for imageId,image in self.items():
      if not image.isDockerImageThere():
        keysToDelete.append(imageId)
    for key in keysToDelete:
      del self[key]



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#external imports
import sys
import optparse
#internal imports
import subuserlib.classes.user
import subuserlib.resolve
import subuserlib.commandLineArguments
import subuserlib.profile
import subuserlib.classes.installationTask
import subuserlib.print

def parseCliArgs(realArgs):
  usage = "usage: subuser print-dependency-info IMAGE_NAME(s) SETS_OF_IMAGES"
  description = """Prints information about how the listed images relate to each other:

Example:

    $ subuser print-dependency-info foo@default
"""
  parser = optparse.OptionParser(usage=usage,description=description,formatter=subuserlib.commandLineArguments.HelpFormatterThatDoesntReformatDescription())
  return parser.parse_args(args=realArgs)

@subuserlib.profile.do_cprofile
def runCommand(realArgs):
  """
  Print the dependencies of the listed progam sources.
  """
  (options,imageSourceIds) = parseCliArgs(realArgs)
  if len(imageSourceIds) == 0:
    sys.exit("No images specified.  Issue this command with -h to view help.")
  user = subuserlib.classes.user.User()
  for imageSourceId in imageSourceIds:
    try:
      imageSource = subuserlib.resolve.resolveImageSource(user,imageSourceId)
    except KeyError:
      sys.exit("The image "+imageSourceId+" does not exist.")
    indent = 0
    for imageSourceInLineage in subuserlib.classes.installationTask.getTargetLineage(imageSource):
      displayLine = (" "*indent) + imageSourceInLineage.getIdentifier()
      subuserlib.print.printWithoutCrashing(displayLine)
      indent = indent + 1



# -*- coding: utf-8 -*-

"""
Contains code that prepairs a subuser's image to be run.
"""

#external imports
import os
#internal imports
from subuserlib.classes.userOwnedObject import UserOwnedObject
import subuserlib.classes.exceptions
import subuserlib.docker

class RunReadyImage(UserOwnedObject):
  def __init__(self,user,subuser):
    self.__subuser = subuser
    self.__id = None
    UserOwnedObject.__init__(self,user)

  def setup(self):
    if not "run-ready-image-id" in self.getSubuser().getRuntimeCache():
      self.__id = self.build()
      self.getSubuser().getRuntimeCache()["run-ready-image-id"] = self.__id
      self.getSubuser().getRuntimeCache().save()

  def getSubuser(self):
    return self.__subuser

  def getId(self):
    if not self.__id:
      self.__id = self.getSubuser().getRuntimeCache()["run-ready-image-id"]
    return self.__id

  def generateImagePreparationDockerfile(self):
    """
    There is still some preparation that needs to be done before an image is ready to be run.  But this preparation requires run time information, so we cannot preform that preparation at build time.
    """
    dockerfileContents  = "FROM "+self.getSubuser().getImageId()+"\n"
    dockerfileContents += "RUN useradd --uid="+str(self.getUser().getEndUser().uid)+" "+self.getUser().getEndUser().name+" ;export exitstatus=$? ; if [ $exitstatus -eq 4 ] ; then echo uid exists ; elif [ $exitstatus -eq 9 ]; then echo username exists. ; else exit $exitstatus ; fi\n"
    dockerfileContents += "RUN test -d "+self.getUser().getEndUser().homeDir+" || mkdir "+self.getUser().getEndUser().homeDir+" && chown "+self.getUser().getEndUser().name+" "+self.getUser().getEndUser().homeDir+"\n"
    if self.getSubuser().getPermissions()["serial-devices"]:
      dockerfileContents += "RUN groupadd dialout; export exitstatus=$? ; if [ $exitstatus -eq 4 ] ; then echo gid exists ; elif [ $exitstatus -eq 9 ]; then echo groupname exists. ; else exit $exitstatus ; fi\n"
      dockerfileContents += "RUN groupadd uucp; export exitstatus=$? ; if [ $exitstatus -eq 4 ] ; then echo gid exists ; elif [ $exitstatus -eq 9 ]; then echo groupname exists. ; else exit $exitstatus ; fi\n"
      dockerfileContents += "RUN usermod -a -G dialout "+self.getUser().getEndUser().name+"\n"
      dockerfileContents += "RUN usermod -a -G uucp "+self.getUser().getEndUser().name+"\n"
    if self.getSubuser().getPermissions()["sudo"]:
      dockerfileContents += "RUN (umask 337; echo \""+self.getUser().getEndUser().name+" ALL=(ALL) NOPASSWD: ALL\" > /etc/sudoers.d/allowsudo )\n"
    return dockerfileContents

  def build(self):
    """
    Returns the Id of the Docker image to be run.
    """
    try:
      tag = subuserlib.docker.buildImageTag("subuser-"+self.getUser().getEndUser().name+"-"+self.getSubuser().getName(),self.getSubuser().getPermissions().getHash())
      return self.getUser().getDockerDaemon().build(None,quietClient=True,useCache=True,tag=tag,forceRm=True,rm=True,dockerfile=self.generateImagePreparationDockerfile())
    except subuserlib.classes.exceptions.ImageBuildException as e:
      self.getUser().getRegistry().log("Error building run-ready image for subuser "+self.getSubuser().getName()+"\n"+str(e))
      return None



# -*- coding: utf-8 -*-

"""
Stores metadata about images which are built to encorporate changes to subuser images which are required in order to implement various permissions.
"""

#external imports
import os
import json
#internal imports
from subuserlib.classes.userOwnedObject import UserOwnedObject
from subuserlib.classes.fileBackedObject import FileBackedObject

class RuntimeCache(dict,UserOwnedObject,FileBackedObject):
  def __init__(self,user,subuser):
    self.__subuser = subuser
    UserOwnedObject.__init__(self,user)
    self.load()

  def getPathToCurrentImagesRuntimeCacheDir(self):
    return os.path.join(self.getUser().getConfig()["runtime-cache"],self.getSubuser().getImageId())

  def getRuntimeCacheFilePath(self):
    return os.path.join(self.getPathToCurrentImagesRuntimeCacheDir(),self.getSubuser().getPermissions().getHash()+".json")

  def getSubuser(self):
    return self.__subuser

  def save(self):
    try:
      self.getUser().getEndUser().makedirs(self.getPathToCurrentImagesRuntimeCacheDir())
    except OSError:
      pass
    with self.getUser().getEndUser().get_file(self.getRuntimeCacheFilePath(),mode='w') as runtimeCacheFileHandle:
      json.dump(self,runtimeCacheFileHandle,indent=1,separators=(',',': '))

  def reload(self):
    self.save()
    self.load()

  def load(self):
    if not self.getSubuser().getImageId():
      raise NoRuntimeCacheForSubusersWhichDontHaveExistantImagesException("No runnable image for subuser "+self.getSubuser().getName()+" found. Use\n\n $ subuser repair\n\nTo repair your instalation.")
    runtimeCacheFilePath = self.getRuntimeCacheFilePath()
    if os.path.exists(runtimeCacheFilePath):
      with open(runtimeCacheFilePath,mode="r") as runtimeCacheFileHandle:
        runtimeCacheInfo = json.load(runtimeCacheFileHandle)
        self.update(runtimeCacheInfo)

class NoRuntimeCacheForSubusersWhichDontHaveExistantImagesException(Exception):
  pass



# -*- coding: utf-8 -*-

"""
Container objects allow you to interact with docker containers via the Docker daemon.
"""

#external imports
import json,urllib
#internal imports
from subuserlib.classes.userOwnedObject import UserOwnedObject

class Container(UserOwnedObject):
  def __init__(self,user,containerId):
    self.__containerId = containerId
    UserOwnedObject.__init__(self,user)

  def inspect(self):
    """
     Returns a dictionary of container properties.
     If the container no longer exists, return None.
    """
    self.getUser().getDockerDaemon().getConnection().request("GET","/v1.13/containers/"+self.__containerId+"/json")
    response = self.getUser().getDockerDaemon().getConnection().getresponse()
    if not response.status == 200:
      response.read() # Read the response and discard it to prevent the server from getting locked up: http://stackoverflow.com/questions/3231543/python-httplib-responsenotready
      return None
    else:
      return json.loads(response.read().decode("utf-8"))

  def stop(self):
    self.getUser().getDockerDaemon().getConnection().request("POST","/v1.13/containers/"+self.getId()+"/stop")
    response = self.getUser().getDockerDaemon().getConnection().getresponse()
    response.read()

  def remove(self,force=False):
    queryParameters =  {
      'force': force
      }
    try:
      queryParametersString = urllib.urlencode(queryParameters)
    except AttributeError:
      queryParametersString = urllib.parse.urlencode(queryParameters) # Python 3
    self.getUser().getDockerDaemon().getConnection().request("DELETE","/v1.13/containers/"+self.getId()+"?"+queryParametersString)
    response = self.getUser().getDockerDaemon().getConnection().getresponse()
    response.read()

  def getId(self):
    return self.__containerId



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#external imports
import sys
import optparse
#internal imports
from subuserlib.classes.user import User
import subuserlib.verify
import subuserlib.commandLineArguments
import subuserlib.profile
from subuserlib.classes.permissionsAccepters.acceptPermissionsAtCLI import AcceptPermissionsAtCLI

####################################################
def parseCliArgs(realArgs):
  usage = "usage: subuser repair [options]"
  description = """
Repair your subuser installation.

This is usefull when migrating from one machine to another.  You can copy your ~/.subuser folder to the new machine and run repair, and things should just work.
"""
  parser = optparse.OptionParser(usage=usage,description=description,formatter=subuserlib.commandLineArguments.HelpFormatterThatDoesntReformatDescription())
  parser.add_option("--accept",dest="accept",action="store_true",default=False,help="Acceppt permissions without asking.")
  parser.add_option("--prompt",dest="prompt",action="store_true",default=False,help="Prompt before installing new images.")
  return parser.parse_args(args=realArgs)

@subuserlib.profile.do_cprofile
def runCommand(realArgs):
  options,arguments=parseCliArgs(realArgs)
  user = User()
  user.getRegistry().commit_message = " ".join(["subuser","repair"]+realArgs)
  permissionsAccepter = AcceptPermissionsAtCLI(user,alwaysAccept = options.accept)
  with user.getRegistry().getLock() as LockFileHandle:
    subusers = user.getRegistry().getSubusers().getSortedList()
    subuserlib.verify.verify(user,subusers=subusers,permissionsAccepter=permissionsAccepter,prompt=options.prompt)
    user.getRegistry().commit()



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#external imports
import sys
import optparse
#internal imports
from subuserlib.classes.user import User
import subuserlib.resolve
import subuserlib.repository
import subuserlib.commandLineArguments
import subuserlib.profile

def parseCliArgs(sysargs):
  usage = "usage: subuser repository [options] [add|remove] NAME <URL>"
  description = """Add or remove a new named repository.

- EXAMPLE
    Add a new repository named foo with the URI http://www.example.com/repo.git.

    $ subuser repository add foo http://www.example.com/repo.git
    $ #You can also add a local repository:
    $ subuser repository add local-foo file:///home/timothy/my-local-repo/

- EXAMPLE
    Remove the repository named foo.

    $subuser repository remove foo

  """
  parser=optparse.OptionParser(usage=usage,description=description,formatter=subuserlib.commandLineArguments.HelpFormatterThatDoesntReformatDescription())
  return parser.parse_args(args=sysargs)

@subuserlib.profile.do_cprofile
def runCommand(sysargs):
  """
  Manage named subuser repositories.
  """
  options,args = parseCliArgs(sysargs)
  user = User()
  user.getRegistry().commit_message = " ".join(["subuser","repository"]+sysargs)
  try:
    action = args[0]
  except IndexError:
    sys.exit("Use subuser repository --help for help.")
  if action == "add":
    if not len(args) == 3:
      sys.exit("Use subuser repository --help for help.")
    name = args[1]
    url = args[2]
    with user.getRegistry().getLock():
      subuserlib.repository.add(user,name,url)
  elif action == "remove":
    if not len(args) == 2:
      sys.exit("Use subuser repository --help for help.")
    name = args[1]
    with user.getRegistry().getLock():
      subuserlib.repository.remove(user,name)
  else:
     sys.exit("Action "+args[0]+" not supported. Please see:\n subuser repository --help")



#!/usr/bin/python3
import sys,subprocess
sys.path.append("../../logic")
import subuserlib.commands

builtInCommands = subuserlib.commands.getBuiltIn()

for command in builtInCommands:
  with open(command+".rst","w") as command_file:
    commandDocs = command + "\n"+("="*len(command))+"\n"
    print("Collecting help for:"+command)
    commandHelpOutput = subprocess.check_output(["../../logic/subuser",command,"--help"])
    commandHelpOutput= commandHelpOutput.decode()
    commandDocs += commandHelpOutput.replace("\n\n    $","\n::\n\n    $")
    command_file.write(commandDocs)

with open("index.rst",mode="w") as index:
  with open("index.rst.head",mode="r") as indexHead:
    index.write(indexHead.read())
  for command in builtInCommands:
    index.write("  "+command+"\n")
  index.write("\n")



# -*- coding: utf-8 -*-
# Taken from: http://www.openpanel.com/2007/08/http-on-unix-sockets-with-python/

"""
This module allows us to communicate using http over standard unix sockets.
"""

#external imports
import socket
try:
  import httplib
except ImportError:
  import http.client
  httplib = http.client
#internal imports
#import ...


class UHTTPConnection(httplib.HTTPConnection):
  """
  Subclass of Python library HTTPConnection that uses a unix-domain socket.
  """
  def __init__(self, path):
    httplib.HTTPConnection.__init__(self, 'localhost')
    self.path = path

  def connect(self):
    sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
    sock.connect(self.path)
    self.sock = sock



# -*- coding: utf-8 -*-

"""
Module used for the loading and saving of permissions.json files. Contains the default permissions list.
"""

#external imports
import json
import collections
import sys
import os
#internal imports
# import ...
allImagesMustHavePermissions = "All subuser images must have a permissions.json file as defined by the permissions.json standard: <http://subuser.org/subuser-standard/permissions-dot-json-file-format.html>"

# Defaults from http://subuser.org/subuser-standard/permissions-dot-json-file-format.html
# This is a comprehensive list of all permissions
defaults = {
 # Conservative permissions
  "description": ""
 ,"maintainer": ""
 ,"executable": None
 ,"basic-common-permissions": False
 ,"stateful-home": False
 ,"inherit-locale": False
 ,"inherit-timezone": False
 ,"entrypoints":{}
 # Moderate permissions
 ,"gui": None
 ,"user-dirs": []
 ,"inherit-envvars": []
 ,"sound-card": False
 ,"webcam": False
 ,"access-working-directory": False
 ,"allow-network-access": False
 # Liberal permissions
 ,"x11": False
 ,"system-dirs": {}
 ,"graphics-card": False
 ,"serial-devices":False
 ,"system-dbus": False
 ,"as-root": False
 ,"sudo": False
 # Anarchistic permissions
 ,"privileged": False
 ,"run-commands-on-host": False
 }

guiDefaults = {
 "clipboard": False,
 "system-tray": False,
 "cursors": False,
 "border-color": "red"
 }

basicCommonPermissions = ["stateful-home","inherit-locale","inherit-timezone"]

levels = [
  {"name" : "prelude",
   "description" : "",
   "permissions" : ["description", "maintainer", "executable","entrypoints"]},
  {"name" : "conservative",
   "description" : "Conservative permissions(These are safe):",
   "permissions" : ["stateful-home", "inherit-locale", "inherit-timezone"]},
  {"name" : "moderate",
   "description" : "Moderate permissions(These are probably safe):",
   "permissions" : ["gui", "user-dirs", "sound-card", "webcam", "access-working-directory", "allow-network-access"]},
  {"name" : "liberal",
   "description" : "Liberal permissions(These may pose a security risk):",
   "permissions" : ["x11", "system-dirs", "graphics-card", "serial-devices", "system-dbus", "as-root"]},
  {"name" : "anarchistic",
   "description" : "WARNING: These permissions give the subuser full access to your system when run.",
   "permissions" : ["privileged","run-commands-on-host"]}]

descriptions = {
  # Prelude
  "description":lambda description : ["Description: "+description]
  ,"maintainer":lambda maintainer : ["Maintainer: "+maintainer]
  ,"executable":lambda executable : ["Executable: "+executable] if executable else ["Is a library."]
  ,"entrypoints": lambda entrypoints: ["Entry points: '"+ "' '".join(entrypoints.keys())+"'"] if entrypoints else []
  # Conservative
  ,"stateful-home": lambda p : ["To have its own home directory where it can save files and settings."] if p else []
  ,"inherit-locale": lambda p : ["To find out which language you speak and what region you live in."] if p else []
  ,"inherit-timezone": lambda p : ["To find out your current timezone."] if p else []
  # Moderate
  ,"gui": lambda guiOptions : (["To be able to display windows."] + sum([guiDescriptions[permission](value) for permission,value in guiOptions.items()],[])) if guiOptions else []
  ,"user-dirs": lambda userDirs : ["To access to the following user directories: '~/"+"' '~/".join(userDirs)+"'"] if userDirs else []
  ,"inherit-envvars": lambda envVars : ["To access to the following environment variables: "+" ".join(envVars)] if envVars else []
  ,"sound-card": lambda p : ["To access to your soundcard, can play sounds/record sound."] if p else []
  ,"webcam": lambda p : ["To access your computer's webcam/can see you."] if p else []
  ,"access-working-directory": lambda p: ["To access the directory from which it was launched."] if p else []
  ,"allow-network-access": lambda p : ["To access the network/internet."] if p else []
  # Liberal
  ,"x11": lambda p : ["To display X11 windows and interact with your X11 server directly(log keypresses, read over your shoulder, steal your passwords, control your computer ect.)"] if p else []
  ,"system-dirs": lambda systemDirs : ["To read and write to the host's `"+source+"` directory, mounted in the container as:`"+dest+"`" for source,dest in systemDirs.items()]
  ,"graphics-card": lambda p : ["To access your graphics-card directly for OpenGL tasks."] if p else []
  ,"serial-devices": lambda p : ["To access serial devices such as programmable micro-controllers and modems."] if p else []
  ,"system-dbus": lambda p: ["To talk to the system dbus daemon."] if p else []
  ,"sudo": lambda p: ["To be allowed to use the sudo command to run subproccesses as root in the container."] if p else []
  ,"as-root": lambda p: ["To be allowed to run as root within the container."] if p else []
  # Anarchistic
  ,"privileged": lambda p: ["To have full access to your system.  To even do things as root outside of its container."] if p else []
  ,"run-commands-on-host": lambda p: ["To run commands as a normal user on the host system."] if p else []
  }

guiDescriptions = {
  "clipboard": lambda p : ["Is able to access the host's clipboard."] if p else []
  ,"system-tray": lambda p : ["Is able to create system tray icons."] if p else []
  ,"cursors": lambda p : ["Is able to change the mouse's cursor icon."] if p else []
  ,"border-color": lambda p : ["Window borders will be "+p] if p else ["Window borders will be red."]}

def load(permissionsFilePath=None,permissionsString=None):
  """
  Return a dictionary of permissions from the given permissions.json file.  Permissions that are not specified are set to their default values.

  Loading permissions doesn't crash horribly and does something at least slightly sensible.

  >>> getNonDefaultPermissions(load(permissionsString="{}"))
  OrderedDict()

  Valid paths are loaded when the user-dirs permission is set.

  >>> print(json.dumps(getNonDefaultPermissions(load(permissionsString='{\"user-dirs\":[\"Downloads\"]}'))))
  {"user-dirs": ["Downloads"]}

  However, funny business is strictly frowned upon.

  >>> getNonDefaultPermissions(load(permissionsString='{\"user-dirs\":[\"..\"]}'))
  Traceback (most recent call last):
    ...
  SyntaxError: Error, user dir permissions may not contain relative paths. The user dir: ".." is forbidden.

  >>> getNonDefaultPermissions(load(permissionsString='{\"user-dirs\":[\"../\"]}'))
  Traceback (most recent call last):
    ...
  SyntaxError: Error, user dir permissions may not contain relative paths. The user dir: "../" is forbidden.

  >>> getNonDefaultPermissions(load(permissionsString='{\"user-dirs\":[\"./../../\"]}'))
  Traceback (most recent call last):
    ...
  SyntaxError: Error, user dir permissions may not contain relative paths. The user dir: "./../../" is forbidden.

  Even if it is well hiden behind valid values.

  >>> getNonDefaultPermissions(load(permissionsString='{\"user-dirs\":[\"Downloads\",\"../\"]}'))
  Traceback (most recent call last):
    ...
  SyntaxError: Error, user dir permissions may not contain relative paths. The user dir: "../" is forbidden.

  >>> getNonDefaultPermissions(load(permissionsString='{\"user-dirs\":[\"Downloads\",\"./Foo\"]}'))
  Traceback (most recent call last):
    ...
  SyntaxError: Error, user dir permissions may not contain relative paths. The user dir: "./Foo" is forbidden.

  >>> getNonDefaultPermissions(load(permissionsString='{\"user-dirs\":[\"Downloads\",\"/Foo\"]}'))
  Traceback (most recent call last):
    ...
  SyntaxError: Error, user dir permissions may not contain system wide absolute paths. The user dir: "/Foo" is forbidden.

  >>> getNonDefaultPermissions(load(permissionsString='{\"x11\":true,\"gui\":{}}'))
  Traceback (most recent call last):
    ...
  SyntaxError: X11 and GUI permissions are mutually exclusive. Cannot provide direct and indirect access to X11 at the same time.

  """
  if not permissionsString is None:
    try:
      permissions=json.loads(permissionsString, object_pairs_hook=collections.OrderedDict)
    except ValueError:
      sys.exit(permissionsString+" is not valid json.  "+allImagesMustHavePermissions)
  else:
    # read permissions.json file
    with open(permissionsFilePath, 'r') as file_f:
      try:
        permissions=json.load(file_f, object_pairs_hook=collections.OrderedDict)
      except ValueError:
        raise SyntaxError(permissionsFilePath+" is not valid json.  "+allImagesMustHavePermissions)
  # Validate that the permissions are supported by this version of subuser.
  for permission in permissions.keys():
    if not permission in defaults:
      raise SyntaxError("Error: the permission \""+permission+"\" is not supported by your version of subuser.  Try updating first.")
    else: # Validate if the permissions don't have unsafe contents.
      if permission == "user-dirs":
        userDirs = permissions["user-dirs"]
        if userDirs and type(userDirs) is list:
          for userDir in permissions["user-dirs"]:
            if ".." in userDir or userDir.startswith("./"):
              raise SyntaxError("Error, user dir permissions may not contain relative paths. The user dir: \""+userDir+"\" is forbidden.")
            if userDir.startswith("/"):
              raise SyntaxError("Error, user dir permissions may not contain system wide absolute paths. The user dir: \""+userDir+"\" is forbidden.")
        elif userDirs is []:
          SyntaxError("The user-dirs permission must be a list of directory names.")
  if "gui" in permissions and not permissions["gui"] is None:
    for guiPermission in permissions["gui"].keys():
      if not guiPermission in guiDefaults:
        raise SyntaxError("Error: the gui permission \""+guiPermission+"\" is not supported by your version of subuser.  Try updating first.")
  # Set permission defaults for permissions that are not explicitly specified in the permissions.json file
  if "basic-common-permissions" in permissions and permissions["basic-common-permissions"]:
    for basicCommonPermission in basicCommonPermissions:
      if not basicCommonPermission in permissions:
        permissions[basicCommonPermission] = True
  for permission,defaultValue in defaults.items():
    if not permission in permissions:
      permissions[permission] = defaultValue
  # gui permissions
  if not permissions["gui"] is None:
    for permission,defaultValue in guiDefaults.items():
      if not permission in permissions["gui"]:
        permissions["gui"][permission] = defaultValue
    if permissions["x11"]:
      raise SyntaxError("X11 and GUI permissions are mutually exclusive. Cannot provide direct and indirect access to X11 at the same time.")
  return permissions

def getNonDefaultPermissions(permissions):
  """
  Returns the dictionary of permissions which are NOT set to their default values.

  >>> import subuserlib.permissions
  >>> permissions = subuserlib.permissions.load(permissionsString='{"x11":true}')
  >>> getNonDefaultPermissions(permissions) == {u'x11': True}
  True
  """
  nonDefaultPermissions = collections.OrderedDict()
  for permission,value in permissions.items():
    if not value == defaults[permission]:
      nonDefaultPermissions[permission] = value
  return nonDefaultPermissions

def getJSONString(permissions):
  """
  Returns the given permissions as a JSON formated string.
  """
  permissionsToSave = getNonDefaultPermissions(permissions)
  return json.dumps(permissionsToSave,indent=1, separators=(',', ': '))

def compare(oldDefaults,newDefaults,userApproved):
  """
  Analize permission sets for changes.
  First, compare the old defaults to the user approved permissions.
  By doing so, we aquire an understanding of which permissions have been set by the user, and which are still at their default values.
  We will leave the user-set permissions alone.

  Next, we compair the old non-user set permissions to the new defaults.

  Finally, we return a list of permissions that have been removed as well as a dictionary of permissions which have been added or changed.

  The return value is a tuple of the form: ([removed-permisions],{additions/changes})
  """
  return __compare(oldDefaults = getNonDefaultPermissions(oldDefaults), newDefaults = getNonDefaultPermissions(newDefaults), userApproved = getNonDefaultPermissions(userApproved))

def __compare(oldDefaults,newDefaults,userApproved):
  """
  Analize permission sets for changes.
  First, compare the old defaults to the user approved permissions.
  By doing so, we aquire an understanding of which permissions have been set by the user, and which are still at their default values.
  We will leave the user-set permissions alone.

  Next, we compair the old non-user set permissions to the new defaults.

  Finally, we return a list of permissions that have been removed as well as a dictionary of permissions which have been added or changed.

  >>> import subuserlib.permissions
  >>> subuserlib.permissions.__compare(oldDefaults={"a":1,"b":2,"c":3,"d":4,"e":5},newDefaults={"a":1,"b":3,"c":4,"f":4},userApproved={"a":1,"b":5,"e":5,"z":7}) == (["e"],{"f":4})
  True
  """
  userSetPermissions = {}
  for key,value in userApproved.items():
    # Cleaver code is evil, but I have given into temptation in this case ^_^
    if (not key in oldDefaults) or (oldDefaults[key] != value):
      userSetPermissions[key] = value
  for key,value in oldDefaults.items():
    if (not key in userApproved):
      userSetPermissions[key] = value
  addedOrChangedPermissions = {}
  for key,value in newDefaults.items():
    #                                                         added  or  changed
    if (not key in userSetPermissions) and ((not key in oldDefaults) or (key in oldDefaults and oldDefaults[key] != value)):
      addedOrChangedPermissions[key] = value
  droppedPermissions = []
  for key in oldDefaults.keys():
    if (key not in userSetPermissions) and (key not in newDefaults):
      droppedPermissions.append(key)
  return (droppedPermissions,addedOrChangedPermissions)



# -*- coding: utf-8 -*-
"""
This module helps us figure out which subuser subcommands can be called.
"""

#external imports
import os
import subprocess
import sys
import pkgutil
#internal imports
import subuserlib.executablePath
import subuserlib.paths
import subuserlib.builtInCommands

def getBuiltIn():
  """
  Get a list of the names of the built in subuser commands.
  """
  # Ugly hack courtesy of http://stackoverflow.com/questions/487971/is-there-a-standard-way-to-list-names-of-python-modules-in-a-package
  return [name for _, name, _ in pkgutil.iter_modules([os.path.dirname(subuserlib.builtInCommands.__file__)])]

def getExternal():
  """
  Return the list of "external" subuser commands.  These are not built in commands but rather stand alone executables which appear in the user's $PATH and who's names start with "subuser-"
  """
  def isPathToCommand(path):
    directory, executableName = os.path.split(path)
    return executableName.startswith("subuser-")
  externalCommandPaths = subuserlib.executablePath.queryPATH(isPathToCommand,list=True)
  externalCommands = []
  subuserPrefixLength=len("subuser-")
  for externalCommandPath in externalCommandPaths:
    commandDir, executableName = os.path.split(externalCommandPath)
    commandName = executableName[subuserPrefixLength:]
    if commandName.endswith(".py"):
      commandName=commandName[:-3]
    externalCommands.append(commandName)
  return list(set(externalCommands)) # remove duplicate entries

def getCommands():
  """
  Returns a list of commands that may be called by the user.
  """
  return list(set(getBuiltIn() + getExternal()))

def getCommand(command):
  if command in getBuiltIn():
    commandModule = __import__("subuserlib.builtInCommands."+command, fromlist=[''])
    return commandModule.runCommand
  else:
    externalCommandPath = subuserlib.executablePath.which("subuser-"+command)
    if externalCommandPath:
      def runCommand(args):
        sys.exit(subprocess.call([externalCommandPath]+args))
      return runCommand
    else:
      return None



# -*- coding: utf-8 -*-

"""
High level operations on subusers.
"""

#external imports
import sys
import os
import shutil
#internal imports
import subuserlib.resolve
import subuserlib.classes.subuser
import subuserlib.verify
import subuserlib.update
import subuserlib.classes.exceptions as exceptions

def add(user,subuserName,imageSourceIdentifier,permissionsAccepter,prompt=False,forceInternal=False):
  if subuserName.startswith("!") and not forceInternal:
    sys.exit("A subusers may not have names beginning with ! as these names are reserved for internal use.")
  if subuserName in user.getRegistry().getSubusers():
    sys.exit("A subuser named "+subuserName+" already exists.")
  user.getRegistry().logChange("Adding subuser "+subuserName+" with image "+imageSourceIdentifier)
  try:
    imageSource = subuserlib.resolve.resolveImageSource(user,imageSourceIdentifier)
  except (KeyError,subuserlib.resolve.ResolutionError) as keyError:
    sys.exit("Could not add subuser.  The image source "+imageSourceIdentifier+" does not exist.\n"+str(keyError))
  addFromImageSource(user,subuserName,imageSource,permissionsAccepter,prompt)

def addFromImageSource(user,subuserName,imageSource,permissionsAccepter,prompt=False):
  addFromImageSourceNoVerify(user,subuserName,imageSource)
  subuser = user.getRegistry().getSubusers()[subuserName]
  subuserlib.verify.verify(user,subusers=[subuser],permissionsAccepter=permissionsAccepter,prompt=prompt)
  user.getRegistry().commit()

def addFromImageSourceNoVerify(user,subuserName,imageSource):
  subuser = subuserlib.classes.subuser.Subuser(user,subuserName,None,False,False,[],imageSource=imageSource)
  user.getRegistry().getSubusers()[subuserName] = subuser

def remove(user,subusers):
  didSomething = False
  for subuser in subusers:
    user.getRegistry().logChange("Removing subuser "+str(subuser.getName()))
    try:
      subuserHome = subuser.getHomeDirOnHost()
      if subuserHome and os.path.exists(subuserHome):
        user.getRegistry().logChange(" If you wish to remove the subusers home directory, issule the command $ rm -r "+subuserHome)
    except:
      pass
    user.getRegistry().logChange(" If you wish to remove the subusers image, issue the command $ subuser remove-old-images")
    for serviceSubuserName in subuser.getServiceSubuserNames():
      try:
        serviceSubuser = user.getRegistry().getSubusers()[serviceSubuserName]
        serviceSubuser.removePermissions()
        del user.getRegistry().getSubusers()[serviceSubuserName]
      except KeyError:
        pass
    # Remove service locks
    try:
      shutil.rmtree(os.path.join(user.getConfig()["lock-dir"],"services",subuser.getName()))
    except OSError:
      pass
    # Remove permission files
    subuser.removePermissions()
    del user.getRegistry().getSubusers()[subuser.getName()]
    didSomething = True
  if didSomething:
    subuserlib.verify.verify(user)
    user.getRegistry().commit()

def setExecutableShortcutInstalled(user,subusers,installed):
  for subuser in subusers:
    if installed:
      user.getRegistry().logChange("Adding launcher for subuser "+subuser.getName()+" to $PATH.")
    else:
      user.getRegistry().logChange("Removing launcher for subuser "+subuser.getName()+" from $PATH.")
    subuser.setExecutableShortcutInstalled(installed)
  subuserlib.verify.verify(user)
  user.getRegistry().commit()

def setEntrypointsExposed(user,subusers,exposed,permissionsAccepter):
  for subuser in subusers:
    if exposed:
      user.getRegistry().logChange("Exposing entrypoints for subuser "+subuser.getName()+" in the $PATH.")
    else:
      user.getRegistry().logChange("Removing entrypoints for subuser "+subuser.getName()+" from $PATH.")
    subuser.setEntrypointsExposed(exposed)
  subuserlib.verify.verify(user,subusers=subusers,permissionsAccepter=permissionsAccepter)
  user.getRegistry().commit()



# -*- coding: utf-8 -*-

"""
This is a PermissionAccepter object used to get user approval of permissions via the command line.
"""

#external imports
from collections import OrderedDict
#internal imports
from subuserlib.classes.permissionsAccepters.permissionsAccepter import PermissionsAccepter
from subuserlib.classes.userOwnedObject import UserOwnedObject
import subuserlib.permissions
import subuserlib.print

class AcceptPermissionsAtCLI(PermissionsAccepter,UserOwnedObject):
  def __init__(self,user,alwaysAccept = False):
    self.__alwaysAccept = alwaysAccept
    UserOwnedObject.__init__(self,user)

  def getAllwaysAccept(self):
    """
    Should the accepter accept the permissions/changes without actually prompting the users?
    """
    return self.__alwaysAccept

  def accept(self,subuser,newDefaults,oldDefaults,userApproved):
    if userApproved is None:
      subuserlib.print.printWithoutCrashing(subuser.getName()+u": would like to have the following permissions:")
      newDefaults.describe()
      createNewPermissions = True
    else:
      createNewPermissions = False
      (removedPermissions,additionsAndChanges) = subuserlib.permissions.compare(newDefaults = newDefaults, oldDefaults=oldDefaults, userApproved=userApproved)
      if additionsAndChanges == {} and removedPermissions == [] and not subuser.wereEntryPointsExposedThisRun():
        return
      if not additionsAndChanges == {}:
        subuserlib.print.printWithoutCrashing(subuser.getName()+" would like to add/change the following permissions:")
        for permission,value in additionsAndChanges.items():
          for line in subuserlib.permissions.descriptions[permission](value):
            subuserlib.print.printWithoutCrashing("   - "+line)
      if not removedPermissions == []:
        subuserlib.print.printWithoutCrashing("")
        subuserlib.print.printWithoutCrashing(subuser.getName()+" no longer needs the following permissions:")
        for removedPermission in removedPermissions:
          for line in subuserlib.permissions.descriptions[removedPermission](oldDefaults[removedPermission]):
            subuserlib.print.printWithoutCrashing("   - "+line)
      if "entrypoints" in additionsAndChanges and subuser.areEntryPointsExposed():
        subuserlib.print.printWithoutCrashing(subuser.getName()+" would like to expose the following entrypoints to the system PATH:")
        for entrypoint in additionsAndChanges["entrypoints"].keys():
          subuserlib.print.printWithoutCrashing(entrypoint)
      if subuser.wereEntryPointsExposedThisRun():
        if subuser.getPermissions()["entrypoints"]:
          subuserlib.print.printWithoutCrashing(subuser.getName()+" would like to expose the following entrypoints to the system PATH:")
          for entrypoint in subuser.getPermissions()["entrypoints"].keys():
            subuserlib.print.printWithoutCrashing(entrypoint)
        else:
          subuserlib.print.printWithoutCrashing("Entrypoints marked to be exposed, but nothing to expose.")
          if additionsAndChanges == {} and removedPermissions == []:
            return
    options = OrderedDict([("A","Accept and apply changes")
                          ,("E","Apply changes and edit result")
                          ,("e","Ignore request and edit permissions by hand")
                          ,("r","Reject permissions.")])
    if createNewPermissions:
      del options["e"]
    for option,description in options.items():
      subuserlib.print.printWithoutCrashing(option+" - "+description)
    if self.getAllwaysAccept():
      subuserlib.print.printWithoutCrashing("A")
      choice = "A"
    else:
      choice = None
    while not choice in options:
      try:
        choice = input("Please select an option:")
      except EOFError:
        choice = "r"
    if (choice == "A") or (choice == "E"):
      if createNewPermissions:
        subuser.createPermissions(newDefaults)
      else:
        subuser.getPermissions().applyChanges(removedPermissions,additionsAndChanges)
      subuser.getPermissions().save()
    if (choice == "E") or (choice == "e"):
      subuser.editPermissionsCLI()
    if choice == "r":
      if createNewPermissions:
        subuser.createPermissions(subuserlib.permissions.load(permissionsString="{}"))
        subuser.getPermissions().save()



# -*- coding: utf-8 -*-
# pylint: disable=no-init,old-style-class

"""
A file structure object is a read only view of a file structure, such as a real directory or a git tree.
"""

#external imports
import abc
import hashlib
import os
import stat
import sys
#internal imports
import subuserlib.test
if subuserlib.test.testing:
  hashtestDir = subuserlib.test.hashtestDir

class FileStructure():
  __metaclass__ = abc.ABCMeta

  @abc.abstractmethod
  def ls(self, subfolder):
    """
    Returns a list of file and folder paths.
    Paths are relative to the root of the FileStructure.
    """
    pass

  @abc.abstractmethod
  def lsFiles(self,subfolder):
    """
    Returns a list of paths to files in the subfolder.
    Paths are relative to the root of the FileStructure.
    """
    pass

  @abc.abstractmethod
  def lsFolders(self,subfolder):
    """
    Returns a list of paths to folders in the subfolder.
    Paths are relative to the root of the FileStructure.
    """
    pass

  @abc.abstractmethod
  def exists(self,path):
    pass

  @abc.abstractmethod
  def read(self,path):
    """
    Returns the contents of the given file.
    """
    pass

  @abc.abstractmethod
  def readBinary(self,path):
    """
    Returns the contents of the given file.
    """
    pass

  @abc.abstractmethod
  def getMode(self,path):
    pass

  def getModeString(self,path):
    """
    Return the human readable mode string for the mode in octal notation.
    """
    octalMode = oct(self.getMode(path))
    if sys.version_info[0] == 2:
      octalMode = octalMode[1:]
    if sys.version_info[0] == 3:
      octalMode = octalMode[2:]
    return octalMode

  def hash(self,path):
    """
    Return the SHA1 hash of the file or directory.

    In the case of directectories, hashes alphabetically and hashes subdirectories first.

    Hashes the following:
      - Relative file path
      - File mode
      - File contents

    Return the hash as a hexidecimal string.

    >>> from subuserlib.classes.fileStructure import FileStructure
    >>> fileStructure = BasicFileStructure(subuserlib.classes.fileStructure.hashtestDir)
    >>> fileStructure.hash("./")
    '6b9c28475016167ba6b58ad37ea9eb56d7364cb9'
    """
    SHAhash = hashlib.sha1()
    # TODO - what about symlinks?
    # TODO - what about devices?
    # TODO - what about hard link loops?
    # TODO - what about sockets?
    # TODO - what about named pipes?
    def hashFile(path):
      encodedPath = path.encode("utf-8","replace")
      SHAhash.update(encodedPath)
      SHAhash.update(self.getModeString(path).encode("utf-8"))
      SHAhash.update(self.readBinary(path))
    def hashDir(path):
      # Hash subdirectories
      subdirs = self.lsFolders(path)
      subdirs.sort()
      for subdir in subdirs:
        hashDir(subdir)
      # Hash files
      files = self.lsFiles(path)
      files.sort()
      for fileToHash in files:
        hashFile(fileToHash)
    hashDir(path)
    return SHAhash.hexdigest()

class BasicFileStructure(FileStructure):
  """
  A FileStructure backed by real files.
  """
  def __init__(self,path):
    self.__path = path

  def getPath(self):
    return self.__path

  def getPathInStructure(self,path):
    """
    Given a relative path within the file structure, return an absolute path.
    """
    return os.path.join(self.getPath(),path)

  def ls(self, subfolder):
    """
    >>> from subuserlib.classes.fileStructure import FileStructure
    >>> fileStructure = BasicFileStructure(subuserlib.classes.fileStructure.hashtestDir)
    >>> print(",".join(fileStructure.ls("./")))
    bar,blah
    """
    paths = []
    path = self.getPathInStructure(subfolder)
    for path in os.listdir(path):
      paths.append(os.path.normpath(os.path.join(subfolder,path)))
    paths.sort()
    return paths

  def lsFiles(self,subfolder):
    """
    >>> from subuserlib.classes.fileStructure import FileStructure
    >>> import os
    >>> fileStructure = BasicFileStructure(subuserlib.classes.fileStructure.hashtestDir)
    >>> print(",".join(fileStructure.lsFiles("./")))
    blah
    """
    files = []
    for path in self.ls(subfolder):
      if os.path.isfile(self.getPathInStructure(path)):
        files.append(os.path.normpath(path))
    return files

  def lsFolders(self,subfolder):
    """
    >>> from subuserlib.classes.fileStructure import FileStructure
    >>> import os
    >>> fileStructure = BasicFileStructure(subuserlib.classes.fileStructure.hashtestDir)
    >>> print(",".join(fileStructure.lsFolders("./")))
    bar
    """
    folders = []
    for path in self.ls(subfolder):
      pathInStructure = self.getPathInStructure(path)
      if os.path.isdir(pathInStructure):
        folders.append(os.path.normpath(path))
    return folders

  def exists(self,path):
    """
    >>> from subuserlib.classes.fileStructure import FileStructure
    >>> import os
    >>> fileStructure = BasicFileStructure(subuserlib.classes.fileStructure.hashtestDir)
    >>> fileStructure.exists("./blah")
    True
    >>> fileStructure.exists("./non-existant")
    False
    """
    return os.path.exists(self.getPathInStructure(path))

  def read(self,path):
    """
    >>> from subuserlib.classes.fileStructure import BasicFileStructure
    >>> import os
    >>> fileStructure = BasicFileStructure(subuserlib.classes.fileStructure.hashtestDir)
    >>> print(fileStructure.read("./blah"))
    blahblah
    <BLANKLINE>
    """
    with open(self.getPathInStructure(path),"r",encoding="utf-8") as fd:
      return fd.read()

  def readBinary(self,path):
    """
    >>> from subuserlib.classes.fileStructure import FileStructure
    >>> import os
    >>> fileStructure = BasicFileStructure(subuserlib.classes.fileStructure.hashtestDir)
    >>> print(fileStructure.read("./blah"))
    blahblah
    <BLANKLINE>
    """
    with open(self.getPathInStructure(path),"rb") as fd:
      return fd.read()

  def getMode(self,path):
    """
    >>> from subuserlib.classes.fileStructure import FileStructure
    >>> import os
    >>> fileStructure = BasicFileStructure(subuserlib.classes.fileStructure.hashtestDir)
    >>> print(fileStructure.getModeString("./blah"))
    100664
    """
    return os.stat(self.getPathInStructure(path))[stat.ST_MODE]



#!/usr/bin/python3
import setuptools
import os
import inspect

def readme():
  with open("README.rst") as f:
    return f.read()

def version():
  with open("VERSION") as f:
    return f.read().strip()

pathToThisSourceFile = os.path.abspath(inspect.getfile(inspect.currentframe()))

commands = ["logic/subuser","logic/execute-json-from-fifo"]

setuptools.setup(
  name="subuser",
  version=version(),
  description="subuser - a program which helps you run other programs in containers, securely and portably.",
  long_description=readme(),
  classifiers=[
    "Development Status :: 3 - Alpha",
    "License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)",
    "Programming Language :: Python :: 3.0",
    ],
  keywords="subuser Docker containers security portability",
  url="http://subuser.org",
  author="Timothy Hobbs",
  author_email="timothy@hobbs.cz",
  license="LGPLv3",
  packages=["subuserlib","subuserlib.classes","subuserlib.builtInCommands","subuserlib.classes.docker","subuserlib.classes.subuserSubmodules","subuserlib.classes.subuserSubmodules.run","subuserlib.classes.permissionsAccepters"],
  package_dir={"subuserlib": "logic/subuserlib",
               "subuserlib.builtInCommands":"logic/subuserlib/builtInCommands",
               "subuserlib.classes":"logic/subuserlib/classes",
               "subuserlib.classes.docker":"logic/subuserlib/classes/docker",
               "subuserlib.classes.subuserSubmodules":"logic/subuserlib/classes/subuserSubmodules",
               "subuserlib.classes.subuserSubmodules.run":"logic/subuserlib/classes/subuserSubmodules/run",
               "subuserlib.classes.permissionsAccepters":"logic/subuserlib/classes/permissionsAccepters"},
  package_data={'subuserlib': ['data/*']},
  scripts=commands,
  include_package_data=True,
  zip_safe=False)



# -*- coding: utf-8 -*-

"""
This is one of the most important modules in subuser.  This module has one function `verify` which is used to apply the changes for most commands that change the user's configuration:

 - Adding a new subuser
 - Removing an old subuser
 - Updating repositories
 - Adding repositories
 - Repairing the installation
"""

#external imports
import shutil
import os
#internal imports
from subuserlib.classes.installationTask import InstallationTask
import subuserlib.classes.exceptions as exceptions
import subuserlib.classes.subuser

def verify(user,permissionsAccepter=None,checkForUpdatesExternally=False,subusers=[],prompt=False,useCache=False):
  """
   Ensure that:
     - Registry is consistent; warns the user about subusers that point to non-existant source images.
     - For each subuser there is an up-to-date image installed.
     - No-longer-needed temporary repositories are removed. All temporary repositories have at least one subuser who's image is built from one of the repository's image sources.
     - No-longer-needed installed images are removed.
  """
  user.getRegistry().log("Verifying subuser configuration.")
  user.getRegistry().log("Verifying registry consistency...")
  for subuser in subusers:
    try:
      subuser.getImageSource()
    except subuserlib.classes.subuser.NoImageSourceException:
      user.getRegistry().log("WARNING: "+subuser.getName()+" is no longer present in it's source repository. Support for this progam may have been dropped.")
      try:
        subusers.remove(subuser)
      except ValueError:
        pass
  user.getRegistry().log("Unregistering any non-existant installed images.")
  user.getInstalledImages().unregisterNonExistantImages()
  user.getRegistry().cleanOutOldPermissions()
  if subusers:
    user.getRegistry().setChanged(True)
    user.getRegistry().log("Approving permissions...",verbosityLevel=3)
    (failedSubusers,permissionParsingExceptions) = approvePermissions(user,subusers,permissionsAccepter)
    user.getRegistry().log("Permissions approved...",verbosityLevel=3)
    subusers = [x for x in subusers if x not in failedSubusers]
    for failedSubuser in failedSubusers:
      try:
        failedSubuser.getPermissions()
      except subuserlib.classes.subuser.SubuserHasNoPermissionsException:
        del user.getRegistry().getSubusers()[failedSubuser.getName()]
    subusers += ensureServiceSubusersAreSetup(user,subusers)
    installationTask = InstallationTask(user,subusersToBeUpdatedOrInstalled=subusers,checkForUpdatesExternally=checkForUpdatesExternally)
    outOfDateSubusers = installationTask.getOutOfDateSubusers()
    if outOfDateSubusers:
      user.getRegistry().log("New images for the following subusers need to be installed:")
      for subuser in outOfDateSubusers:
        user.getRegistry().log(subuser.getName())
      if (not prompt) or (prompt and (not input("Would you like to install those images now? [Y/n]") == "n")):
        installationTask.updateOutOfDateSubusers(useCache=useCache)
    for exception in permissionParsingExceptions:
      user.getRegistry().log(str(exception))
    subusersWhosImagesFailedToBuild = installationTask.getSubusersWhosImagesFailedToBuild()
    if subusersWhosImagesFailedToBuild:
      user.getRegistry().log("Images for the following subusers failed to build:")
    for subuser in subusersWhosImagesFailedToBuild:
      user.getRegistry().log(subuser.getName())
    for subuser in subusers:
      try:
        subuser.getRunReadyImage().setup()
        subuser.setupHomeDir()
      except subuserlib.classes.subuserSubmodules.run.runtimeCache.NoRuntimeCacheForSubusersWhichDontHaveExistantImagesException:
        pass
  user.getInstalledImages().save()
  trimUnneededTempRepos(user)
  rebuildBinDir(user)
  cleanupRuntimeDirs(user)
  cleanUpRuntimeCache(user)
  cleanUpAfterImproperlyTerminatedServices(user)

def approvePermissions(user,subusers,permissionsAccepter):
  subusersWhosPermissionsFailedToParse = []
  exceptions = []
  for subuser in subusers:
    if subuser.locked():
      continue
    try:
      userApproved = subuser.getPermissions()
    except subuserlib.classes.subuser.SubuserHasNoPermissionsException:
      userApproved = None
    try:
      oldDefaults = subuser.getPermissionsTemplate()
      newDefaults = subuser.getImageSource().getPermissions()
      permissionsAccepter.accept(subuser=subuser,oldDefaults=oldDefaults,newDefaults=newDefaults,userApproved=userApproved)
      subuser.getPermissionsTemplate().update(subuser.getImageSource().getPermissions())
      subuser.getPermissionsTemplate().save()
    except SyntaxError as e:
      subusersWhosPermissionsFailedToParse.append(subuser)
      exceptions.append(e)
    except subuserlib.classes.subuser.NoImageSourceException:
      subusersWhosPermissionsFailedToParse.append(subuser)
      user.getRegistry().log("Warning: The image source for subuser %s is no longer available."%subuser.getName())
  return (subusersWhosPermissionsFailedToParse,exceptions)

def ensureServiceSubusersAreSetup(user,subusers):
  newServiceSubusers = []
  for subuser in subusers:
    if not subuser.getPermissions()["gui"] is None:
      newServiceSubusers += subuser.getX11Bridge().setup()
  return newServiceSubusers

def trimUnneededTempRepos(user):
  user.getRegistry().log("Running garbage collector on temporary repositories...")
  reposToRemove = []
  for repoId,repo in user.getRegistry().getRepositories().userRepositories.items():
    if repo.isTemporary() and not repo.isInUse():
      user.getRegistry().logChange("Removing uneeded temporary repository: "+repo.getDisplayName())
      repo.removeGitRepo()
      reposToRemove.append(repoId)
  for repoId in reposToRemove:
    del user.getRegistry().getRepositories().userRepositories[repoId]

def rebuildBinDir(user):
  if os.path.exists(user.getConfig()["bin-dir"]):
    shutil.rmtree(user.getConfig()["bin-dir"])
  user.getEndUser().mkdir(user.getConfig()["bin-dir"])
  for _,subuser in user.getRegistry().getSubusers().items():
    if subuser.isExecutableShortcutInstalled():
      subuser.installExecutableShortcut()
    if subuser.areEntryPointsExposed():
      subuser.exposeEntrypoints()

def cleanupRuntimeDirs(user):
  """
  Remove left overs that were not properly cleaned up after running subusers.
  """
  def is_process_running(process_id):
    """
    Taken from: http://stackoverflow.com/questions/7647167/check-if-a-process-is-running-in-python-in-linux-unix
    """
    try:
      os.kill(process_id, 0)
      return True
    except OSError:
      return False
  def clearPIDSubdirs(pidDir):
    """
    Clear out a directory containing subdirectories named with the PIDs of processes by removing any directories corresponding to non-running processes.
    """
    user.getRegistry().log("Clearing directory "+ pidDir)
    try:
      for pid in os.listdir(pidDir):
        try:
          numericPid = int(pid)
          if not is_process_running(numericPid):
            shutil.rmtree(os.path.join(pidDir,pid))
            user.getRegistry().log("Removing "+ os.path.join(pidDir,pid)+" process is no longer running.",verbosityLevel=3)
          else:
            user.getRegistry().log("Not removing "+ os.path.join(pidDir,pid)+" process is still running.",verbosityLevel=3)
        except ValueError:
          pass
    except OSError:
      pass
  # Clean up ~/.subuser/volumes/execute
  clearPIDSubdirs(os.path.join(user.getConfig()["volumes-dir"],"execute"))
  # Clean up ~/.subuser/volumes/x11
  clearPIDSubdirs(os.path.join(user.getConfig()["volumes-dir"],"x11"))

def cleanUpRuntimeCache(user):
  """
  Remove runtime cache directories for no longer existant images.
  """
  runtimeCacheDir = user.getConfig()["runtime-cache"]
  try:
    for imageId in os.listdir(runtimeCacheDir):
      if not imageId in user.getInstalledImages():
        shutil.rmtree(os.path.join(runtimeCacheDir,imageId))
  except FileNotFoundError:
    pass

def cleanUpAfterImproperlyTerminatedServices(user):
  """
  Remove service lock files and service volumes in the case that a service has terminated improperly.
  """
  # Go through xpra volumes.
  try:
    xpraVolumeDir = os.path.join(user.getConfig()["volumes-dir"],"xpra")
    serviceVolumes = os.listdir(xpraVolumeDir)
    serviceLocksDir = os.path.join(user.getConfig()["lock-dir"],"services")
    serviceLockDirs = os.listdir(serviceLocksDir)
    subusersWithServiceDirs = set(serviceLockDirs) | set(serviceVolumes)
  except OSError:
    subusersWithServiceDirs = []
  for subuserWithServiceDirs in subusersWithServiceDirs:
    user.getRegistry().log("Removing left over service files for subuser "+ subuserWithServiceDirs)
    if not subuserWithServiceDirs in user.getRegistry().getSubusers():
      try:
        shutil.rmtree(os.path.join(xpraVolumeDir,subuserWithServiceDirs))
      except OSError as e:
        print(e)
      try:
        shutil.rmtree(os.path.join(serviceLocksDir,subuserWithServiceDirs))
      except OSError as e:
        print(e)
    else:
      user.getRegistry().getSubusers()[subuserWithServiceDirs].getX11Bridge().cleanUpIfNotRunning()



# -*- coding: utf-8 -*-
"""
This module contains configurations for C{optparse}.
"""

#external imports
import optparse
#internal imports
#import ...

class HelpFormatterThatDoesntReformatDescription (optparse.HelpFormatter):
  """
  Format help with indented section bodies but don't reformat the description.
  """
  def __init__(self, indent_increment=2, max_help_position=24, width=None,short_first=1):
    optparse.HelpFormatter.__init__(self, indent_increment, max_help_position, width, short_first)

  def format_usage(self, usage):
    return optparse._("Usage: %s\n") % usage

  def format_heading(self, heading):
    return "%*s%s:\n" % (self.current_indent, "", heading)

  def format_description(self, description):
    return description



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#external imports
import sys
import optparse
import json
import os
import copy
#internal imports
from subuserlib.classes.user import User
import subuserlib.commandLineArguments
import subuserlib.permissions
from subuserlib.classes.permissionsAccepters.acceptPermissionsAtCLI import AcceptPermissionsAtCLI
from subuserlib.classes.permissions import Permissions
import subuserlib.profile
import subuserlib.test

def parseCliArgs(realArgs):
  usage = "usage: subuser pkg [init|add|test] IMAGE-SOURCE-NAMES <args>"
  description = """ Packaging subucommands. These commands are to be issued from the root of a subuser repository.

  init - Initialize a subuser repository.

  add - Add an image source to the repository.
  Usage: $ subuser pkg add image-source-name

  To test images use subuser dev --update <image-name>
  """
  parser=optparse.OptionParser(usage=usage,description=description,formatter=subuserlib.commandLineArguments.HelpFormatterThatDoesntReformatDescription())
  parser.add_option("--accept",dest="accept",action="store_true",default=False,help="When testing images, accept permissions without asking.")
  parser.add_option("--prompt",dest="prompt",action="store_true",default=False,help="Prompt before installing new images.")
  parser.add_option("--image-sources-dir", dest="imageSourcesDir",default=None,help="When initializing a new repository, set where the image source directories will be stored. This may be set when you are working with an existing git repository which you would like to unobtrusively add subuser image sources too.")
  parser.add_option("--permissions-file", dest="permissionsFile",default=None,help="When adding a new image source, the path to the new permissions.json file for the image source. Note: This flag should not be used to import permissions files from outside of the repository. Use the POSIX cp command for that.")
  parser.add_option("--image-file", dest="imageFile",default=None,help="When adding a new image source, the path to the new image file which will be used to build the image source. Note: This flag should not be used to import build files from outside of the repository. Use the POSIX cp command for that.")
  parser.add_option("--build-context", dest="buildContext",default=None,help="When adding a new image source, the path to the build context for building the image.")
  return parser.parse_args(args=realArgs)

defaultImageFileTemplate = """FROM-SUBUSER-IMAGE libx11@default
RUN apt-get update && apt-get upgrade -y && apt-get install -y PKG"""
defaultPermissions = copy.deepcopy(subuserlib.permissions.defaults)
if subuserlib.test.testing:
  defaultImageFileTemplate = """FROM-SUBUSER-IMAGE foo@default"""
  defaultPermissions["executable"] = "/usr/bin/nothing"

@subuserlib.profile.do_cprofile
def runCommand(realArgs):
  """
  Run packaging subcommands.
  """
  options,args = parseCliArgs(realArgs)
  try:
    subcommand = args[0]
    subcommands = ["init","add"]
    if not subcommand in subcommands:
      sys.exit(subcommand + " not a recognized subcommand of pkg. Use --help for help.")
  except IndexError:
    sys.exit("No command given, use -h for help.")
  user = User()
  if subcommand == "init":
    if not os.path.exists("./.subuser.json"):
      repoConfig = {}
      if options.imageSourcesDir:
        repoConfig["image-sources-dir"] = options.imageSourcesDir
      with user.getEndUser().get_file("./.subuser.json","w") as subuserDotJson:
        json.dump(repoConfig,subuserDotJson)
        if options.imageSourcesDir:
          user.getEndUser().makedirs(options.imageSourcesDir)
      subuserlib.print.printWithoutCrashing("Subuser repository initialized successfully!")
      subuserlib.print.printWithoutCrashing("You can add new image sources with:")
      subuserlib.print.printWithoutCrashing("$ subuser pkg add image-source-name")
    else:
      sys.exit("Subuser repository already initialized. Exiting.")
  if subcommand == "add":
    repo = subuserlib.resolve.getRepositoryFromPath(user,os.environ["PWD"])
    imageSourceToAdd = args[1]
    if imageSourceToAdd in repo:
      sys.exit("An image source named "+imageSourceToAdd+" is already present in the repository. Cannot add. Exiting.")
    subuserlib.print.printWithoutCrashing("Adding new image source "+imageSourceToAdd)
    useDefaultLocations = options.imageFile is None and options.permissionsFile is None and options.buildContext is None
    if useDefaultLocations:
      imageSourceDir = os.path.join(repo.getImageSourcesDir(),imageSourceToAdd)
      buildContext = os.path.join(imageSourceDir,"image")
      imageFile = os.path.join(buildContext,"SubuserImagefile")
      permissionsFile = os.path.join(imageSourceDir,"permissions.json")
      try:
        user.getEndUser().makedirs(buildContext)
      except OSError:
        pass
    else:
      if repo.getRepoConfig() is None:
        sys.exit("You must initialize your repository with 'pkg init' before adding image sources to it.")
      if options.buildContext is None or options.imageFile is None or options.permissionsFile is None:
        sys.exit("If you specify non-default paths you must specify all of them. That is --image-file, --build-context AND --permissions-file. Cannot add image. Exiting...")
      imageFile = options.imageFile
      try:
        user.getEndUser().makedirs(os.path.dirname(imageFile))
      except OSError:
        pass
      buildContext = options.buildContext
      try:
        user.getEndUser().makedirs(buildContext)
      except OSError:
        pass
      permissionsFile = options.permissionsFile
      try:
        user.getEndUser().makedirs(os.path.dirname(permissionsFile))
      except OSError:
        pass
      repoConfig = repo.getRepoConfig()
      if not "explicit-image-sources" in repoConfig:
        repoConfig["explicit-image-sources"] = {}
      repoConfig["explicit-image-sources"][imageSourceToAdd] = {"image-file":imageFile,"build-context":buildContext,"permissions-file":permissionsFile}
      with user.getEndUser().get_file("./.subuser.json","w") as subuserDotJson:
        json.dump(repoConfig,subuserDotJson,indent=1,separators=(",",": "))
    permissions = defaultPermissions
    (returncode,maintainerName,stderr) = user.getEndUser().callCollectOutput(["git","config","user.name"])
    subuserlib.print.printWithoutCrashing(stderr)
    (returncode,maintainerEmail,stderr) = user.getEndUser().callCollectOutput(["git","config","user.email"])
    subuserlib.print.printWithoutCrashing(stderr)
    permissions["maintainer"] = maintainerName.rstrip("\n")+" <"+maintainerEmail.rstrip("\n")+">"
    if not os.path.exists(permissionsFile):
      with user.getEndUser().get_file(permissionsFile,"w") as pf:
        json.dump(permissions,pf,indent=1,separators=(",",": "))
    while True:
      user.getEndUser().runEditor(permissionsFile)
      try:
        Permissions(user,initialPermissions=subuserlib.permissions.load(permissionsFilePath=permissionsFile),writePath=permissionsFile).save()
        break
      except SyntaxError as e:
        input(str(e)+"\nPress ENTER to edit the file again.")
    if not os.path.exists(imageFile):
      with user.getEndUser().get_file(imageFile,"w") as imgf:
        imgf.write(defaultImageFileTemplate)
    user.getEndUser().runEditor(imageFile)
    subuserlib.print.printWithoutCrashing("Your application has now been packaged. Run subuser dev %s to test your work. Use subuser dev --update %s to iterate."%(imageSourceToAdd,imageSourceToAdd))



# -*- coding: utf-8 -*-

"""
This is an abstract class providing common methods shared by all service types.

The semantics of the lock file mechanism may be found `here <https://github.com/subuser-security/subuser/issues/31>`_ .
"""

#external imports
import abc
import json
import os
import errno
import fcntl
import shutil
#internal imports
from subuserlib.classes.userOwnedObject import UserOwnedObject

class Service(UserOwnedObject):
  __metaclass__ = abc.ABCMeta

  def __init__(self,user,subuser):
    self.__subuser = subuser
    UserOwnedObject.__init__(self,user)

  @abc.abstractmethod
  def start(self,serviceStatus):
    """
    Start the service. Block untill the service has started. Returns a modified service status dictionary with any service specific properties set.
    """
    pass

  @abc.abstractmethod
  def stop(self,serviceStatus):
    """
    Stop the service. Block untill the service has stopped.
    """
    pass

  @abc.abstractmethod
  def cleanUp(self):
    pass

  @abc.abstractmethod
  def isRunning(self,serviceStatus):
    """
    Returns True if the services is running.
    """
    pass

  @abc.abstractmethod
  def getName(self):
    pass

  def getLockfileDir(self):
    return os.path.join(self.getUser().getConfig()["lock-dir"],"services",self.__subuser.getName())

  def getLockfilePath(self):
    return os.path.join(self.getLockfileDir(),self.getName()+".json")

  def removeLockFile(self):
    os.remove(self.getLockfilePath())
    try:
      os.rmdir(self.getLockfileDir())
    except OSError:
      pass

  def getLock(self):
    try:
      self.getUser().getEndUser().makedirs(self.getLockfileDir())
    except OSError as exception:
      if exception.errno != errno.EEXIST:
        raise
    while True:
      try:
        lockFd = open(self.getLockfilePath(),mode="r+")
        break
      except IOError:
        self.getUser().getEndUser().create_file(self.getLockfilePath())
    fcntl.flock(lockFd,fcntl.LOCK_EX)
    return lockFd

  def addClient(self):
    """
    Increase the services client counter, starting the service if necessary. Blocks untill the service is ready to accept the new client.
    """
    with self.getLock() as lockFile:
      try:
        serviceStatus = json.load(lockFile)
      except ValueError:
        serviceStatus = {}
        serviceStatus["client-counter"] = 0
      if serviceStatus["client-counter"] == 0 or not self.isRunning(serviceStatus):
        serviceStatus["client-counter"] = 0
        serviceStatus = self.start(serviceStatus)
      serviceStatus["client-counter"] = serviceStatus["client-counter"] + 1
      lockFile.seek(0)
      lockFile.truncate()
      json.dump(serviceStatus,lockFile)
      fcntl.flock(lockFile,fcntl.LOCK_UN)

  def removeClient(self):
    """
    Decrease the services client counter, stopping the service if no longer necessary.
    """
    noMoreClients = False
    with self.getLock() as lockFile:
      serviceStatus = json.load(lockFile)
      serviceStatus["client-counter"] = serviceStatus["client-counter"] - 1
      if serviceStatus["client-counter"] < 0:
        raise RemoveClientException("The client-counter is already zero. Client cannot be removed!")
      if serviceStatus["client-counter"] == 0:
        self.stop(serviceStatus)
        noMoreClients = True
      lockFile.seek(0)
      lockFile.truncate()
      json.dump(serviceStatus,lockFile)
      if noMoreClients:
        self.removeLockFile()
      fcntl.flock(lockFile,fcntl.LOCK_UN)

  def cleanUpIfNotRunning(self):
    """
    Check if running.
    If not running, run cleanUp method.
    """
    with self.getLock() as lockFile:
      try:
        serviceStatus = json.load(lockFile)
      except ValueError:
        serviceStatus = {}
        serviceStatus["client-counter"] = 0
      if serviceStatus["client-counter"] == 0 or not self.isRunning(serviceStatus):
        self.cleanUp()
        self.removeLockFile()
      fcntl.flock(lockFile,fcntl.LOCK_UN)


class RemoveClientException(Exception):
  pass



# -*- coding: utf-8 -*-

"""
Implements functions involved in building/installing/updating subuser images.
"""

#external imports
import sys
#internal imports
import subuserlib.classes.installedImage
import subuserlib.verify
from subuserlib.classes.userOwnedObject import UserOwnedObject
import subuserlib.classes.exceptions as exceptions

class InstallationTask(UserOwnedObject):
  def __init__(self,user,subusersToBeUpdatedOrInstalled,checkForUpdatesExternally=False):
    UserOwnedObject.__init__(self,user)
    self.__subusersToBeUpdated = subusersToBeUpdatedOrInstalled
    self.__upToDateImageSources = set()
    self.__outOfDateImageSources = set()
    self.__outOfDateSubusers = None
    self.__subusersWhosImagesFailedToBuild = set()
    self.checkForUpdatesExternally = checkForUpdatesExternally

  def getOutOfDateSubusers(self):
    """
    Returns a list of subusers which are out of date or have no InstalledImage associated with them.
    """
    if self.__outOfDateSubusers is None:
      self.getUser().getRegistry().log("Checking if images need to be updated or installed...")
      self.__outOfDateSubusers = set()
      for subuser in self.__subusersToBeUpdated:
        try:
          if (not subuser.locked()) and (not (subuser.getImageSource().getLatestInstalledImage() is None)):
            self.getUser().getRegistry().log("Checking if subuser "+subuser.getName()+" is up to date.")
            for imageSource in getTargetLineage(subuser.getImageSource()):
              if imageSource in self.__upToDateImageSources:
                continue
              if imageSource in self.__outOfDateImageSources:
                upToDate = False
              else:
                upToDate = self.isUpToDate(imageSource)
              if upToDate:
                self.__upToDateImageSources.add(imageSource)
              else:
                self.__outOfDateImageSources.add(imageSource)
                self.__outOfDateSubusers.add(subuser)
                break
          if subuser.getImageSource().getLatestInstalledImage() is None or subuser.getImageId() is None or not subuser.isImageInstalled():
            if subuser.locked():
              self.getUser().getRegistry().log("Subuser "+subuser.getName()+" has no image. But is locked. Marking for installation anyways.")
            self.__outOfDateSubusers.add(subuser)
        except exceptions.ImageBuildException as e:
          self.getUser().getRegistry().log(str(e))
          self.__subusersWhosImagesFailedToBuild.add(subuser)
    return self.__outOfDateSubusers

  def isUpToDate(self,imageSource):
    installedImage = imageSource.getLatestInstalledImage()
    if installedImage is None:
      return False
    if not installedImage.isDockerImageThere():
      return False
    targetLineage = getTargetLineage(imageSource)
    installedLineage = installedImage.getImageLineage()
    if not (len(targetLineage) == len(installedLineage)):
      return False
    sideBySideLineages = zip(installedLineage,targetLineage)
    for installed,target in sideBySideLineages:
      if target in self.__outOfDateImageSources:
        return False
      if not installed.getImageId() == target.getLatestInstalledImage().getImageId():
        return False
    if not installedImage.getImageSourceHash() == imageSource.getHash():
      return False
    if self.checkForUpdatesExternally and installedImage.checkForUpdates():
      return False
    return True

  def updateOutOfDateSubusers(self,useCache=False):
    """
    Install new images for those subusers which are out of date.
    """
    parent = None
    for subuser in self.getOutOfDateSubusers():
      try:
        for imageSource in getTargetLineage(subuser.getImageSource()):
          if imageSource in self.__upToDateImageSources:
            parent = imageSource.getLatestInstalledImage().getImageId()
          elif imageSource in self.__outOfDateImageSources:
            parent = installImage(imageSource,parent=parent,useCache=useCache)
            self.__outOfDateImageSources.remove(imageSource)
            self.__upToDateImageSources.add(imageSource)
          else:
            if not self.isUpToDate(imageSource):
              parent = installImage(imageSource,parent=parent,useCache=useCache)
            else:
              parent = imageSource.getLatestInstalledImage().getImageId()
            self.__upToDateImageSources.add(imageSource)
        if not subuser.getImageId() == parent:
          subuser.setImageId(parent)
          subuser.getUser().getRegistry().logChange("Installed new image <"+subuser.getImageId()+"> for subuser "+subuser.getName())
      except exceptions.ImageBuildException as e:
        self.getUser().getRegistry().log(str(e))
        self.__subusersWhosImagesFailedToBuild.add(subuser)

  def getSubusersWhosImagesFailedToBuild(self):
    return self.__subusersWhosImagesFailedToBuild

def installImage(imageSource,parent,useCache=False):
  """
  Install a image by building the given ImageSource.
  Register the newly installed image in the user's InstalledImages list.
  Return the Id of the newly installedImage.
  """
  imageSource.getUser().getRegistry().logChange("Installing " + imageSource.getName() + " ...")
  imageId = imageSource.build(parent,useCache=useCache)
  imageSource.getUser().getInstalledImages()[imageId] = subuserlib.classes.installedImage.InstalledImage(imageSource.getUser(),imageId,imageSource.getName(),imageSource.getRepository().getName(),imageSource.getHash())
  imageSource.getUser().getInstalledImages().save()
  return imageId

def getTargetLineage(imageSource):
  """
  Return the lineage of the ImageSource, going from its base dependency up to itself.
  """
  sourceLineage = []
  while imageSource:
    sourceLineage.append(imageSource)
    imageSource = imageSource.getDependency()
  return list(reversed(sourceLineage))



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#external imports
import sys
import optparse
#internal imports
import subuserlib.commandLineArguments
from subuserlib.classes.user import User
import subuserlib.update
import subuserlib.profile
from subuserlib.classes.permissionsAccepters.acceptPermissionsAtCLI import AcceptPermissionsAtCLI

#####################################################################################

def parseCliArgs(realArgs):
  usage = "usage: subuser update [options]"
  description = """Update subuser images.

  all
      Updates all subuser images which have been marked as out of date.

  EXAMPLE:
    $ subuser update all

  subusers
      Updates the specified subusers

  EXAMPLE:
    $ subuser update subusers iceweasel git

  lock-subuser-to SUBUSER GIT-COMMIT
      Don't want a subuser to be updated?  No problem, lock it to a given version with this update sub-command.  Use subuser update log to see a list of possible hashes.

  unlock-subuser SUBUSER
      Unlock the subuser and ensure that it is up-to-date.

"""
  parser=optparse.OptionParser(usage=usage,description=description,formatter=subuserlib.commandLineArguments.HelpFormatterThatDoesntReformatDescription())
  parser.add_option("--accept",dest="accept",action="store_true",default=False,help="Accept permissions without asking.")
  parser.add_option("--prompt",dest="prompt",action="store_true",default=False,help="Prompt before installing new images.")
  parser.add_option("--use-cache",dest="useCache",action="store_true",default=False,help="Use the layer cache when building images.")
  return parser.parse_args(args=realArgs)

#################################################################################################

@subuserlib.profile.do_cprofile
def runCommand(realArgs):
  """
  Update your subuser installation.
  """
  options,args = parseCliArgs(realArgs)
  user = User()
  user.getRegistry().commit_message = " ".join(["subuser","update"]+realArgs)
  permissionsAccepter = AcceptPermissionsAtCLI(user,alwaysAccept = options.accept)
  if len(args) < 1:
    sys.exit("No arguments given. Please use subuser update -h for help.")
  elif ["all"] == args:
    with user.getRegistry().getLock():
      subuserlib.update.all(user,permissionsAccepter=permissionsAccepter,prompt=options.prompt,useCache=options.useCache)
  elif "subusers" == args[0]:
    subuserNamesToUpdate = args[1:]
    subusersToUpdate = []
    for subuserName in subuserNamesToUpdate:
      try:
        subusersToUpdate.append(user.getRegistry().getSubusers()[subuserName])
      except KeyError:
        sys.exit("Subuser "+subuserName+" does not exist. Use --help for help.")
    if subusersToUpdate:
      with user.getRegistry().getLock():
        subuserlib.update.subusers(user,subusers=subusersToUpdate,permissionsAccepter=permissionsAccepter,prompt=options.prompt,useCache=options.useCache)
    else:
      sys.exit("You did not specify any subusers to be updated. Use --help for help. Exiting...")
  elif "lock-subuser-to" == args[0]:
    try:
      subuserName = args[1]
      commit = args[2]
    except IndexError:
      sys.exit("Wrong number of arguments.  Expected a subuser name and a commit.  Try running\nsubuser update --help\n for more info.")
    with user.getRegistry().getLock():
      try:
        subuser = user.getRegistry().getSubusers()[subuserName]
      except KeyError:
        sys.exit("Subuser "+subuserName+" does not exist and therefore cannot be locked. Use --help for help.")
      subuserlib.update.lockSubuser(user,subuser=subuser,commit=commit)
  elif "unlock-subuser" == args[0]:
    try:
      subuserName = args[1]
    except IndexError:
      sys.exit("Wrong number of arguments.  Expected a subuser's name. Try running\nsubuser update --help\nfor more information.")
    try:
      subuser = user.getRegistry().getSubusers()[subuserName]
    except KeyError:
      sys.exit("Subuser "+subuserName+" does not exist. Cannot lock. Use --help for help.")
    with user.getRegistry().getLock():
      subuserlib.update.unlockSubuser(user,subuser=subuser,permissionsAccepter=permissionsAccepter,prompt=options.prompt)
  elif len(args) == 1:
    sys.exit(" ".join(args) + " is not a valid update subcommand. Please use subuser update -h for help.")
  else:
    sys.exit(" ".join(args) + " is not a valid update subcommand. Please use subuser update -h for help.")



# -*- coding: utf-8 -*-

"""
File locking helper functions. Originally taken from http://stackoverflow.com/questions/5255220/fcntl-flock-how-to-implement-a-timeout
"""

#external imports
import os
#internal imports
#import ...
import signal, errno
from contextlib import contextmanager
import fcntl

@contextmanager
def wrapTimeout(seconds):
  def timeout_handler(signum, frame):
    pass

  original_handler = signal.signal(signal.SIGALRM, timeout_handler)

  try:
    signal.alarm(seconds)
    yield
  finally:
    signal.alarm(0)
    signal.signal(signal.SIGALRM, original_handler)

def getLock(fd,timeout=None):
  if not timeout is None:
    with wrapTimeout(timeout):
      return getLock(fd)
  else:
    fcntl.flock(fd, fcntl.LOCK_EX)
    return fd



#!/usr/bin/python3
# -*- coding: utf-8 -*-

#external imports
import sys
import optparse
import json
import os
import uuid
import subprocess
#internal imports
import subuserlib.commandLineArguments
import subuserlib.profile
import subuserlib.paths as paths
from subuserlib.classes.user import User

user = User()

def parseCliArgs(realArgs):
  usage = "usage: subuser dev <args> DEV-IMAGE-NAME"
  description = """ Create and run a subuser related to a dev image.
  """
  parser=optparse.OptionParser(usage=usage,description=description,formatter=subuserlib.commandLineArguments.HelpFormatterThatDoesntReformatDescription())
  parser.add_option("--ls",dest="ls",action="store_true",default=False,help="List dev images.")
  parser.add_option("--update",dest="update",action="store_true",default=False,help="Update dev images associated with this folder. Note: This always uses the layer cache. Use subuser update all to update fully without layer caching.")
  parser.add_option("--remove",dest="remove",action="store_true",default=False,help="Remove dev images associated with this folder.")
  parser.add_option("--entrypoint",dest="entrypoint",default=None,help="Use entrypoint instead of default executable.")
  return parser.parse_args(args=realArgs)

@subuserlib.profile.do_cprofile
def runCommand(realArgs):
  options,args = parseCliArgs(realArgs)
  if options.ls:
    subprocess.call([paths.getSubuserExecutable(),"list","available","./"])
    sys.exit()
  devSubuserRegistry = ".subuser-dev"
  devSubusers = {}
  subuserNames = []
  if os.path.exists(devSubuserRegistry):
    with open(devSubuserRegistry,"r") as fd:
      devSubusers = json.load(fd)
  for devSubuser in devSubusers.values():
    subuserNames.append(devSubuser)

  if options.remove:
    subprocess.call([paths.getSubuserExecutable(),"subuser","remove"]+subuserNames)
    sys.exit()

  if options.update:
    if not subprocess.call([paths.getSubuserExecutable(),"update","--use-cache","subusers"]+subuserNames) == 0:
      sys.exit()

  if len(args) != 1:
    if options.update:
      sys.exit()
    sys.exit("Please pass a single dev image name. Use --help for help.")
  devSubuser = None
  devImage = args[0]
  try:
    devSubuser = devSubusers[devImage]
  except KeyError:
    pass
  if devSubuser is None:
    devSubuser = devImage+"@"+os.path.split(os.path.dirname(os.getcwd()+os.sep))[1]+"-"+str(uuid.uuid4())
    if subprocess.call([paths.getSubuserExecutable(),"subuser","add",devSubuser,devImage+"@./"]) == 0:
      devSubusers[devImage] = devSubuser
      with user.getEndUser().get_file(devSubuserRegistry,"w") as fd:
        json.dump(devSubusers,fd)
  if options.entrypoint is None:
    subprocess.call([paths.getSubuserExecutable(),"run",devSubuser])
  else:
    subprocess.call([paths.getSubuserExecutable(),"run","--entrypoint="+options.entrypoint,devSubuser])



# -*- coding: utf-8 -*-
"""
The Config class is used to hold user wide settings.
"""

#external imports
import os
#internal imports
from subuserlib.classes import userOwnedObject
from subuserlib import loadMultiFallbackJsonConfigFile
from subuserlib import paths

class Config(userOwnedObject.UserOwnedObject, dict):
  def __init__(self,user):
    self.__delitem__ = None
    self.__setitem__ = None
    userOwnedObject.UserOwnedObject.__init__(self,user)
    self._loadConfig()

  def _getSubuserConfigPaths(self):
    """ Returns a list of paths to config.json files in order that they should be looked in. """
    configFileInHomeDir = os.path.join(self.getUser().homeDir,".subuser","config.json")
    configFileInEtc = "/etc/subuser/config.json"
    configFileInSubuserDir = paths.getSubuserDataFile("config.json")
    return [configFileInHomeDir,configFileInEtc,configFileInSubuserDir]

  def _expandPathsInConfig(self,config):
    """
    Go through a freshly loaded config file and expand any environment variables in the paths.
    """
    pathsToExpand = [
      "bin-dir"
      ,"registry-dir"
      ,"installed-images-list"
      ,"locked-subusers-path"
      ,"subuser-home-dirs-dir"
      ,"repositories-dir"
      ,"runtime-cache"
      ,"lock-dir"
      ,"volumes-dir"]
    loadMultiFallbackJsonConfigFile.expandPathsInDict(self.getUser().homeDir,pathsToExpand,config)

  def _loadConfig(self):
    """
    Loads the subuser config: a dictionary of settings used by subuser.
    """
    configFileHierarchy = self._getSubuserConfigPaths()
    config = loadMultiFallbackJsonConfigFile.getConfig(configFileHierarchy)
    self._expandPathsInConfig(config)
    for key,entry in config.items():
      self[key] = entry



# -*- coding: utf-8 -*-
# pylint: disable=unused-argument

"""
In order to make our test suit work, we must use a MockDockerDaemon rather than communicating with a real Docker instance.
"""

#external imports
import json
import os
#internal imports
from subuserlib.classes.userOwnedObject import UserOwnedObject
import subuserlib.classes.docker.dockerDaemon
import subuserlib.print

class MockDockerDaemon(UserOwnedObject):
  def __init__(self,user):
    self.images = {}
    self.nextImageId = 1
    self.newId = None
    UserOwnedObject.__init__(self,user)
    self.imagesPath = os.path.join(user.homeDir,"docker/images.json")
    self.__load()
    self.dockerDaemon = subuserlib.classes.docker.dockerDaemon.RealDockerDaemon(user)
    self.connection = MockConnection(self)
    self.dockerDaemon.getConnection = self.getConnection
    self.dockerDaemon.getImageProperties = self.getImageProperties

  def __load(self):
    with open(self.imagesPath,"r") as imagesFile:
      self.images = json.load(imagesFile)

  def __save(self):
    with open(self.imagesPath,"w") as imagesFile:
      json.dump(self.images,imagesFile)

  def getConnection(self):
    return self.connection

  def getImageProperties(self,imageTagOrId):
    """
     Returns a dictionary of image properties, or None if the image does not exist.
    """
    if imageTagOrId in self.images:
      return self.images[imageTagOrId]
    else:
      return None

  def build(self,relativeBuildContextPath=None,repositoryFileStructure=None,useCache=True,rm=True,forceRm=True,quiet=False,quietClient=False,tag=None,dockerfile=None):
    """
    Build a Docker image.  If a the dockerfile argument is set to a string, use that string as the Dockerfile.  Return the newly created images Id or raises an exception if the build fails.
    """
    while str(self.nextImageId) in self.images:
      self.nextImageId = self.nextImageId+1
    self.newId = str(self.nextImageId)
    parent = dockerfile.split("\n")[0].split(" ")[1].rstrip()
    if "debian" in dockerfile:
      parent = ""
    self.images[self.newId] = {"Id":self.newId,"Parent":parent,"Created":str(len(self.images))}
    self.__save()
    self.dockerDaemon.build(relativeBuildContextPath=relativeBuildContextPath,repositoryFileStructure=repositoryFileStructure,useCache=useCache,rm=rm,forceRm=forceRm,quiet=quiet,tag=tag,dockerfile=dockerfile,quietClient=quietClient)
    return self.newId

  def removeImage(self,imageId):
    del self.images[imageId]
    self.__save()

  def getInfo(self):
    return {"Foo":"bar"}

  def execute(self,args,cwd=None,background=False,backgroundSuppressOutput=True,backgroundCollectStdout=False,backgroundCollectStderr=False):
    subuserlib.print.printWithoutCrashing("Execute docker with args: "+str(args))
    subuserlib.print.printWithoutCrashing("Cwd:"+str(cwd))

class MockResponse():
  def __init__(self,mockDockerDaemon):
    self.mockDockerDaemon = mockDockerDaemon
    self.status = 200
    self.body = b"{\"stream\":\"Building"+"→→→".encode("utf-8")+b"\"}\n{\"stream\":\"Building...\"}\n{\"stream\":\"Building...\"}\n{\"stream\":\"Successfully built "+mockDockerDaemon.newId.encode("utf-8")+b"\"}"

  def read(self,bytes=None):
    if bytes:
      value = self.body[:bytes]
      self.body = self.body[bytes:]
      return value
    else:
      return self.body

class MockConnection():
  def __init__(self,mockDockerDaemon):
    self.mockDockerDaemon = mockDockerDaemon
    self.newId = None

  def request(self,method,url,body=None,headers=None):
    pass

  def getresponse(self):
    return MockResponse(self.mockDockerDaemon)



# -*- coding: utf-8 -*-

"""
Helper functions for running foreign executables.
"""

#external imports
import subprocess
import os
import tempfile
#internal imports
import subuserlib.test

def call(args,cwd=None):
  """
  Same as subprocess.call except here you can specify the cwd.

  Returns subprocess's exit code.
  """
  process = subprocess.Popen(args,cwd=cwd)
  (stdout,stderr) = process.communicate()
  return process.returncode

def callBackground(args,cwd=None,suppressOutput=True,collectStdout=False,collectStderr=False):
  """
  Same as subprocess.call except here you can specify the cwd.
  Returns imediately with the subprocess
  """
  stdout = None
  stderr = None

  if suppressOutput:
    devnull = open(os.devnull,"a")
    stdout = devnull
    stderr = devnull

  if collectStdout:
    temp_stdout = tempfile.TemporaryFile(mode="r")
    stdout = temp_stdout.fileno()
  if collectStderr:
    temp_stderr = tempfile.TemporaryFile(mode="r")
    stderr = temp_stderr.fileno()

  process = subprocess.Popen(args,cwd=cwd,stdout=stdout,stderr=stderr,close_fds=True)

  if collectStdout:
    process.stdout_file = temp_stdout
  if collectStderr:
    process.stderr_file = temp_stderr

  return process

def callCollectOutput(args,cwd=None):
  """
  Run the command and return a tuple with: (returncode,the output to stdout as a string,stderr as a string).
  """
  process = subprocess.Popen(args,stdout=subprocess.PIPE,stderr=subprocess.PIPE,cwd=cwd)
  (stdout,stderr) = process.communicate()
  return (process.returncode,stdout.decode("utf-8"),stderr.decode("utf-8"))

